function bg=BackgroundImage(inputnames,outputname,ch,framelist)
% Usage: bg=BackgroundImage(inputnames,[outputname],ch,framelist) 
% Given a movie, background_image calculates the mean pixel values over
% time, returning the result in "bg" and saving it as an image with the
% filename "outputname". If the input file have mutiple image channels,
% "ch" specifies the channel for the background image. "framelist",  
% specifies the frames for the image, for example, [1,3,5]. 
% The movie must be saved in one of the following
% formats:
% - a series of image files with sequential numerical filenames,
% - an image stack in .tif or .gif format,
% - a Norpix .seq movie, or
% - a movie file readable by Matlab's VideoReader function.
% Specify the movie in "inputnames" (e.g., '0*.png' or 'stack.tif' or
% 'movie.avi').  To specify an output format, include an extension in
% "outputname"; '.tif' is used by default. Requires read_seq.m for use with
% .seq movies. 
% Examples: bg=BackgroundImage('Image_0001_0001.raw','bg1r.tif',1,[1:10:1000]) 

% Written 6 April 2010 by Doug Kelley. 
% Fixed error catching of output names w/o extension 4 May 2010.
% Specified uncompressed output explicitly 25 May 2010. 
% Allowed for 16-bit images 17 September 2010. 
% Allowed for color images and files in another directory 13 April 2011. 
% Added image size check and color type check 4 May 2011. 
% Allowed for grayscale images with bitV depth between 8 and 16, 17 May 2011. 
% Made compatible with uncompressed avi movies (using 
% read_uncompressed_avi.m) 20 October 2011. 
% Made compatible with tiff and gif stacks 13 February 2012. 
% Made compatible with .seq movies (using Norpix2MATLAB.m) 16 August 2014.
% Made compatible with VideoReader (instead of read_uncompressed_avi) 13
% September 2014. 
% Updated 4 June 2015 to use read_seq.m instead of Norpix2MATLAB.m, thereby
% achieving compatibility with newer .seq files. 
% Updated 10 Oct 2019 by J. Tithof to handle .raw movies
% Updated 14 May 2022 by Yiming to choose channel number, choose frames for the background image.
% -=- Set defaults -=-----------------------------------------------------
outputnamedefault='background.tif';
formatdefault='.tif'; % for output, if not specified
nocompress={'.hdf','.png','.tif','.tiff'}; % save these formats uncompressed
framelist_default=[]; %The framelist check 
ch_default=1; %the default channel for the raw file is channel 1 


% -=- Parse inputs -=-----------------------------------------------------
if nargin<1
    error(['Usage: bg = ' mfilename '(inputnames,[outputname],[ch],[framelist])'])
end
if ~exist('outputname','var') || isempty(outputname)
    outputname=outputnamedefault;
end
if ~exist('outputname','var') || isempty(outputname)
    outputname=outputnamedefault;
end

if ~exist('framelist','var') || isempty(framelist)
    framelist=framelist_default;
end

if ~exist('ch','var') || isempty(ch)
    ch=ch_default;
end


%Check if there are three channels
% -=- Decide whether avi, stack, or images; set up -=---------------------
[prefix,~,ext]=fileparts(inputnames);
list=dir(inputnames);
if numel(list)==0
    error(['Sorry, cannot find ' inputnames '.']);
end
if numel(list)==1 && ( strcmpi(ext,'.avi') || ...
        strcmpi(ext,'.mp4') || strcmpi(ext,'.mov') || strcmpi(ext,'.mj2') || ...
        strcmpi(ext,'.wmv') || strcmpi(ext,'.mpg') || strcmpi(ext,'.asf') || ...
        strcmpi(ext,'.asx') || strcmpi(ext,'.m4v') || strcmpi(ext,'.ogg') ) % single movie file
    movtype='avi';
    if isempty(which('VideoReader.m')) % check for req'd helper function
        error(['Sorry, reading .avi files requires ' ...
            'VideoReader.m.'])
    end
    vobj=VideoReader(fullfile(prefix,list.name));
    if strcmp(vobj.VideoFormat(1:3),'RGB')
        bitdepth=vobj.BitsPerPixel/3; % RGB
        ct='RGB';
    elseif strcmp(vobj.VideoFormat,'Indexed')
        error('Sorry, indexed color is not supported.')
    else
        bitdepth=vobj.BitsPerPixel; % grayscale
        ct='grayscale';
    end
    Nf=vobj.NumberOfFrames;
elseif strcmpi(ext,'.seq')
    movtype='seq';
    if isempty(which('read_seq.m')) % check for req'd helper function
        error(['Sorry, reading .seq files requires ' ...
            'read_seq.m.'])
    end
    [~,~,movinfo]=read_seq(list.name,0);
    Nf=movinfo.AllocatedFrames;
    bitdepth=movinfo.ImageBitDepthReal;
    ct='grayscale'; % we'll just assume!
elseif numel(list)==1 && ( strcmpi(ext,'.tif') || ...
        strcmpi(ext,'.tiff') || strcmpi(ext,'.gif') ) % single file, looks like an image stack
    movtype='stack';
    movinfo=imfinfo(list.name);
    if numel(movinfo)==1
        error([inputnames ': Only one frame found.'])
    end
    Nf=numel(movinfo);
    bitdepth=movinfo(1).BitDepth;
    ct=movinfo(1).ColorType;
elseif strcmpi(ext,'.raw')
    movtype='raw';
    if isempty(which('read_raw.m')) % check for req'd helper function
        error(['Sorry, reading .raw files requires ' ...
            'read_raw.m.'])
    end
    [~,~,movinfo]=read_raw(list.name,[],[],0); %
    Nf=movinfo.AllocatedFrames;
    bitdepth=movinfo.ImageBitDepthReal;
    ct='grayscale'; % it's converted to gray scale below
else
    movtype='images';
    movinfo=imfinfo(fullfile(prefix,list(1).name));
    Nf=numel(list);
    bitdepth=movinfo.BitDepth;
    ct=movinfo.ColorType;
end
if ~Nf
    error([inputnames ': No frames found.'])
end

%This line specify the framelist for the background image
if ~exist('framelist','var') || isempty(framelist)
    Nf=Nf;
else
    Nf=numel(framelist);
end

%If isempty framne, use Nf(the entire frames) as the frame range
if isempty(framelist)
   framelist=1:Nf;
end

% -=- Sum all frames -=---------------------------------------------------
switch movtype
case('avi')
    d=read(vobj);
    bg0 = sum(d(:,:,:,framelist),4);
        
case('seq')
    d = read_seq(fullfile(prefix,list.name),[0 inf],0);
    bg0 = sum(d,3);  
case('raw')
    d = read_raw(fullfile(prefix,list.name),[],[],0);
    disp(['Using channel' num2str(ch) 'for background image.'])
    bg0 = sum(d{ch}(:,:,framelist),3); 
    
case('stack')
    for ii=1:Nf
        d=imread(fullfile(prefix,list.name),ii);
        if ii==1
            bg0=double(d);
        else
            bg0=bg0+double(d);
        end % if ii==1
    end
case('images')
    for ii=1:Nf
        d=imread(fullfile(prefix,list(ii).name));
        if ii==1
            bg0=double(d);
        else
            bg0=bg0+double(d);
        end % if ii==1
    end
end % switch movtype

% -=- Handle color format & bit depth -=----------------------------------
if strcmp(ct,'grayscale')
    if bitdepth==8
        bg0=uint8(round(bg0/Nf));
    elseif bitdepth<=16
        bg0=uint16(round(bg0/Nf));
    else
        error(['Sorry, grayscale images must have bit depth between ' ...
            '8 and 16.'])
    end
else
    disp('Converting color images to grayscale.')
    bg0=uint8(round(mean(bg0,3)/Nf));
end

% -=- Save file and return output if requested -=-------------------------
[~,~,ext]=fileparts(outputname);
if isempty(ext) % no extension; will append one
    imwrite(bg0,[outputname formatdefault],'compression','none');
elseif any(strcmp(ext,nocompress))
    imwrite(bg0,outputname,'compression','none');
else
    imwrite(bg0,outputname);
end
if nargout>0
    bg=bg0;
end

