function [x,y,t,ang,br,tt] = Particlefinder(inputnames,threshold,framerange,outputname,bground,arealim,chan,noisy)
% Usage: [x,y,t,ang,br,tt] = ParticleFinder(inputnames,threshold,[framerange],[outputname],[bground],[arealim],[chan,[noisy])
% Given a movie of particle motions, ParticleFinder identifies the
% particles, returning their positions, times, and orientations in x, y, t,
% and ang, respectively. The movie must be saved in one of the following
% formats:
% - a series of image files with sequential numerical filenames,
% - an image stack in .tif or .gif format,
% - a ThorLabs .raw movie with accompanying .xml file,
% - a Norpix .seq movie, or
% - a movie file readable by Matlab's VideoReader function
% in "inputnames" (e.g., '0*.png' or 'stack.tif', or 'movie.avi'). 
% The background can be in the form of a tif, or a number can be given to
% create a dynamic background. In this case, the background subtracted
% from the images is based on a series of nearby frames, with that number
% specified by the user.
% If threshold<0, ParticleFinder seeks particles darker than the background by
% at least threshold; if threshold>0, ParticleFinder seeks particles
% brighter than the background by at least abs(threshold); if threshold is
% a two-element array with threshold(1)<0 and threshold(2)>0,
% ParticleFinder seeks particles darker than the background by at least
% abs(threshold(1)) or brighter than the background by at least
% threshold(2). 
% The function outputs a list of times corresponding to all movie frames,
% including those that contained no particles, is returned as "tt". The
% output "br" is true for each particle that was brighter than the,
% background and false for each particle that was darker than the
% background. The background is read from the file "bground"; see
% BackgroundImage. 
% Frames outside the range specified by the two-element
% vector "framerange" are ignored. If arealim==1, ParticleFinder seeks
% single-pixel particles by comparing brightness to adjacent pixels (fast
% and good for small particles); otherwise ParticleFinder seeks particles
% having areas bounded by the two elements of the vector "arealim" (in
% square pixels; this method is better for tracking large particles). If
% "outputname" is not empty, particle positions are also saved as a binary
% file of that name. The file begins with a long int giving the frame
% count, then each frame begins with a long int giving its particle count,
% and continues with double-precision floats giving the x and y coordinates
% of each particle. 
% The channel input, chan, is the channel containing particles, e.i.
% 2 for green particles
% If noisy~=0, the movie is repeated with particle
% locations overlaid. If noisy>1, each movie frame is also saved to disk as
% an image.
% Additionally, if a file called masked_ROI.mat exists and contains
% information about a region of interest, ParticleFinder will only search
% for particles in the previously specified region of interest.
% See also BackgroundImage.m and PredictiveTrackerClass.m.

% Written 20 October 2011 by Doug Kelley, largely based on
% PredictiveTracker.m.
% Renamed ParticleFinder and incorporated FindParticles function 27 October
% 2011.
% Fixed plotting bug in bug plotting 15 November 2011.
% Added "ang" output (particle orientation) 18 November 2011.
% Added "framerange" input 28 November 2011.
% Updated to use weighted centroid 2 December 2011.
% Included invert==-1 option 22 February 2012.
% Made compatible with tiff & gif stacks and squelched regionprops
% divide-by-zero warning 7 March 2012.
% Thanks go to Max Gould for spotting a bug in the main loop; fixed 26
% March 2012.
% Changed 29 March 2012 to plot bars if minarea>1.
% Updated 30 March 2012 to use *weighted* orientaion and pre-allocate
% arrays.
% Updated 29 May 2012 with "arealim" input instead of "minarea".
% Updated 26 December 2013 to use VideoReader instead of
% read_uncompressed_avi; also fixed bugs with arealim.
% Updated 4 September 2014 to properly handle single tiffs or gifs with
% just one frame.
% Updated 13 September 2014 for compatibility with Norpix .seq files.
% Updated 22 October 2014 to give progress updates and label figures when
% working with .seq files.
% Added 'file not found' error message 11 February 2015.
% Updated 4 June 2015 to use read_seq.m instead of Norpix2MATLAB.m, thereby
% achieving compatibility with newer .seq files.
% Added "tt" output 31 October 2017.
% Updated 30 July 2018 to output br, to allow two-element threshold
% input, and to drop "invert" input, since it's redundant.
% Updated 1 August 2018 to allow dynamic backgrounds, using Jeff Tithof's
% code. Also fixed a short-lived bug in computing particle centers in
% FindParticles.
% Updated 25 Sept 2018 by J. Tithof to ensure output variables are column
% vectors. (If Np==0 for the first frame, variables are saved as row
% vectors instead and this was problematic for PredictiveTracker.m.)
% Updated 7 October 2019 by Doug Kelley for compatibility with ThorLabs
% .raw files.
% Updated 10 July 2020 by Doug Kelley to be compatible with (new version
% of?) readFrame.m: frame number is not an input.
% Updated 02 April 2022 by Keelin Quirk to have channel as an input
% Updated 22 April 2022 by Keelin Quirk to isolate a region of interest
% from a file called maskerd_roi.mat if such a file exists
% Updated 12 May 2022 by Keelin Quirk to accept double or uint8 3D matrix 
% as an infile
% Updated 7 September 2022 by Keelin Quirk to read frames from .mov and
% .avi files using read instead of readFrames to fix time indexing issue.
% Fixed numeric background so that it will run for an RBG movie file. 
%Note-numeric background with an RBG movie runs very slowly

% -=- Set defaults -=-----------------------------------------------------
framerange_default = [1 inf]; % by default, all frames
bground_default = 'background.tif';
noisy_default=0; % don't plot unless requested
arealim_default=[1 inf];
pausetime=1/30; % seconds to pause between frames when plotting
savedirname='particlesmovie';
barsize=8; % if interested in angle, plot particles as rods 8 px long
ch=chan; % When reading .raw files, assume particles are on channel 2

assert(nargin>=2,['Usage: [x,y,t,ang,br,tt] = ' mfilename ...
    '(inputnames,threshold,[framerange],[outputname],' ...
    '[bground],[arealim],[noisy])'])
if ~exist('framerange','var') || isempty(framerange)
    framerange=framerange_default;
elseif numel(framerange)==1
    framerange=framerange*[1 1];
end
if ~exist('bground','var') || isempty(bground)
    bground=bground_default;
elseif isnumeric(bground) && ( bground<1 || mod(bground,1) )
    error('If bground is numeric, it must be a nonnegative integer.')
end
if ~exist('arealim','var') || isempty(arealim)
    arealim=arealim_default;
elseif numel(arealim)==1
    arealim=[arealim inf];
end
if ~exist('noisy','var') || isempty(noisy)
    noisy=noisy_default;
end
if ~exist('outputname','var') || isempty(outputname)
    writefile=false;
else
    writefile=true;
end

% -=- Decide whether stack, images, or movie; set up -=---------------------
if ischar(inputnames) == 1
    [filepath,~,ext]=fileparts(inputnames);
    names=dir(inputnames);
    if numel(names)==0
        error(['Sorry, cannot find ' inputnames '.']);
    elseif numel(names)==1 && ( strcmpi(ext,'.tif') || ...
            strcmpi(ext,'.tiff') || strcmpi(ext,'.gif') ) % single tiff or gif probably an image stack
        movtype='stack';
        movinfo=imfinfo(fullfile(filepath,names.name));
        color_depth=2^(movinfo(1).BitDepth);
        ht=movinfo(1).Height;
        wd=movinfo(1).Width;
        tmin=max([framerange(1) 1]);
        tmax=min([framerange(2) numel(movinfo)]);
    elseif strcmpi(ext,'.seq')
        movtype='seq';
        if isempty(which('read_seq.m')) % check for req'd helper function
            error('Sorry, reading .seq files requires read_seq.m.')
        end
        [~,~,movinfo]=read_seq(fullfile(filepath,names.name),0);
        color_depth=2^(movinfo.ImageBitDepthReal);
        ht=movinfo.ImageHeight;
        wd=movinfo.ImageWidth;
        tmin=max([framerange(1) 0]);
        tmax=min([framerange(2) movinfo.AllocatedFrames-1]);
    elseif strcmpi(ext,'.raw')
        movtype='raw';
        if isempty(which('read_raw.m')) % check for req'd helper function
            error('Sorry, reading .raw files requires read_raw.m.')
        end
        [~,~,movinfo]=read_raw(fullfile(filepath,names.name),1);
        color_depth=2^(movinfo.ImageBitDepthReal);
        ht=movinfo.ImageHeight;
        wd=movinfo.ImageWidth;
        tmin=max([framerange(1) 0]);
        tmax=min([framerange(2) movinfo.RecordedFrames-1]);
    elseif numel(names)==1 && ( strcmpi(ext,'.avi') || ...
            strcmpi(ext,'.mp4') || strcmpi(ext,'.mov') || strcmpi(ext,'.mj2') || ...
            strcmpi(ext,'.wmv') || strcmpi(ext,'.mpg') || strcmpi(ext,'.asf') || ...
            strcmpi(ext,'.asx') || strcmpi(ext,'.m4v') || strcmpi(ext,'.ogg') ) % single movie file
        movtype='avi';
        if isempty(which('VideoReader.m')) % check for req'd helper function
            error('Sorry, reading this format requires VideoReader.m.')
        end
        vobj=VideoReader(fullfile(filepath,names.name));
        if strcmp(vobj.VideoFormat(1:3),'RGB')
            color_depth=2^(vobj.BitsPerPixel/3); % RGB
        elseif strcmp(vobj.VideoFormat,'Indexed')
            error('Sorry, indexed color is not supported.')
        else
            color_depth=2^vobj.BitsPerPixel; % grayscale
        end
        ht=vobj.Height;
        wd=vobj.Width;
        tmin=max([framerange(1) 1]);
        tmax=min([framerange(2) round(vobj.Duration*vobj.FrameRate)]); % vobj.NumberOfFrames is not recommended.
        
    else % multiple files; probably a series of images
        movtype='image';
        movinfo=imfinfo(fullfile(filepath,names(1).name)); % assuming first file has same properties as the rest
        color_depth=2^(movinfo.BitDepth);
        ht=movinfo.Height;
        wd=movinfo.Width;
        for ii=1:numel(names)
            if ~isnumeric(bground) && ...
                    strcmp(fullfile(filepath,names(ii).name),bground)
                names(ii)=[]; % don't try to track the background file
                break
            end
        end
        tt=NaN(size(names));
        for ii=1:numel(names)
            [~,myname]=fileparts(names(ii).name);
            tt(ii)=str2double(myname);
        end
        if any(isnan(tt)) % at least one filename was not a number, so just make ordinals
            tt=1:numel(tt);
        end
        tmin=max([framerange(1) min(tt)]);
        tmax=min([framerange(2) max(tt)]);
        if tmax<tmin
            error(['Sorry, ' inputnames ...
                ' does not contain frames in range [' num2str(framerange) ...
                '].'])
        end
        names = names( (tt>=tmin) & (tt<=tmax) );
    end % if numel(names)==0
else %ischar = 0 if array is input
    movtype='array';
    names = inputnames; %for consistency with existing naming
    sz=size(names);
    tt=1:sz(3);
    ht=sz(1);
    wd=sz(2);
    %shorten to desired framerange
    tmin=max([framerange(1) min(tt)]);
    tmax=min([framerange(2) max(tt)]);
        if tmax<tmin
            error(['Sorry, ' inputnames ...
                ' does not contain frames in range [' num2str(framerange) ...
                '].'])
        end
    names = names(:,:, (tt>=tmin) & (tt<=tmax) );
    %if double, need to change max value from 1 to 225
    if isa(names,'double') ==1
        %if not a double, these things will be specified later
        tt=tmin:tmax;
        tt=tt(:);
        Nf=numel(tt); % frame count
        for j=1:Nf
            names(:,:,j) = names(:,:,j).*255;
        end
    end
end
tt=tmin:tmax;
tt=tt(:);
Nf=numel(tt); % frame count

% -=- Pre-compute logarithms for locating particle centers -=-------------
if all(arealim==[1 inf])
    logs = 1:color_depth;
    logs = [log(0.0001) log(logs)];
end

% -=- Generate background image -=----------------------------------------
if isnumeric(bground) % dynamic background
    bgimgs=NaN(ht,wd,bground);
    for ii=1:bground
        t1=tmin+ii-1; % current time
        switch movtype
            case('avi')
                vobj=VideoReader(fullfile(filepath,names.name));
                bggray = double(read(vobj,t1));
                    if strcmp(vobj.VideoFormat(1:3),'RGB')
                        bggray = rgb2gray(bggray);
                    end
                bgimgs(:,:,ii) = bggray;           
            case('seq')
                im = read_seq(fullfile(filepath,names.name),t1,0);
                bgimgs(:,:,ii)=double(im);
            case('raw')
                im = read_raw(fullfile(filepath,names.name),t1,[],0);
                im=im{ch};
                bgimgs(:,:,ii)=double(im);
            case('stack')
                bgimgs(:,:,ii) = ...
                    double(imread(fullfile(filepath,names.name),t1));
            case('images')
                bgimgs(:,:,ii) = ...
                    double(imread(fullfile(filepath,names(ii).name)));
            case('array')
                bgimgs(:,:,ii)=names(:,:,t1);
        end % switch movtype
    end
else % if isnumeric(bground)
    if exist(bground,'file')==2
        background = double(imread(bground));
    else
        warning('MATLAB:ParticleFinder:noBackgroundFile', ...
            ['Cannot find ' bground '. Using blank image instead.'])
        bground='blank background';
        if numel(threshold)==1 && threshold<0 % dark particles
            background = color_depth*ones(ht,wd);
        else
            background = zeros(ht,wd);
        end
    end % if exist(bground,'file')==2
    if ndims(background)==3 % check for RGB instead of grayscale
        background=round(mean(background,3));
    end
end % if isnumeric(bground)

% -=- Set up output file and variables -=---------------------------------
if writefile
    if exist(outputname,'file')
        yn=input(['File ' outputname ' exists. Overwrite (y/n)? '],'s');
        if ~strcmpi(yn(1),'y')
            disp('Not overwritten.')
            writefile=false;
        else
            disp(['Replacing file ' outputname '.'])
            fid=fopen(outputname,'w');
            fwrite(fid,Nf,'int32'); % header is integer frame count
        end
        pause(1)
    else
        fid=fopen(outputname,'w');
        fwrite(fid,Nf,'int32'); % header is integer frame count
    end
end
begins=ones(Nf+1,1);
for ii=1:Nf
    t1=tmin+ii-1; % current time
    switch movtype
        case('avi')
            im = double(read(vobj,t1));
        case('seq')
            im = read_seq(fullfile(filepath,names.name),t1,0);
            im=double(im);
        case('raw')
            im = read_raw(fullfile(filepath,names.name),t1);
            im=double(im{ch});
        case('stack')
            im = double(imread(fullfile(filepath,names.name),t1));
        case('images')
            im = double(imread(fullfile(filepath,names(ii).name)));
        case('array')
            im=names(:,:,t1);
    end % switch movtype
    if ndims(im)==3
        if size(im,3)>3
            error('Sorry, only grayscale and RGB images are supported.')
        end
        im=round(mean(im,3)); % convert to grayscale if necessary
    end

    if isnumeric(bground)
        % Read in another image for the background and replace the oldest
        % image, if we're not at the beginning or the end of the time series
        if ii>round(bground/2) && ii<=Nf-round(bground/2)
            t2=tmin+ii-1+round(bground/2); % current time
            switch movtype
                case('avi')
                    vobj=VideoReader(fullfile(filepath,names.name));
                    im2 = double(read(vobj,t2));
                    if strcmp(vobj.VideoFormat(1:3),'RGB')
                        im2 = rgb2gray(im2);
                    end
                case('seq')
                    im2 = read_seq(fullfile(filepath,names.name),t2,0);
                    im2=double(im2);
                case('raw')
                    im2 = read_raw(fullfile(filepath,names.name),t2);
                    im2=double(im2{ch});
                case('stack')
                    im2 = double(imread(fullfile(filepath,names.name),t2));
                case('images')
                    im2 = double(imread(fullfile(filepath,names(ii).name)));
                case('array')
                    im2=names(:,:,t2);
            end % switch movtype
            bgimgs(:,:,mod(ii-round(bground/2)-1,bground)+1)=im2;
        end
        background=mean(bgimgs,3);
    end % if isnumeric(bground)

    im = round(im - background); % for now, seek any contrast, bright or dark
 
    if exist('masked_ROI.mat') ~= 0
        load('masked_ROI'); %Yiming Add this line to select a region of ROI
        BW=roipoly(im,xroi,yroi); 
        [R,C]=size(BW);
       for i=1:R
            for j=1:C
                if BW(i,j)==1
                    im(i,j)=im(i,j);
                else
                    im(i,j)=0;
                end
            end
       end
    else
    end
    
    if all(arealim==[1 inf])
        [pos,br1] = FindParticles(im,threshold,logs);
    else
        [pos,ang1,br1] = FindRegions(im,threshold,arealim);
    end
    if numel(threshold)==1 && threshold>0 % bright particles
        pos=pos(br1,:); % keep only bright particles
        br1=br1(br1);
    elseif numel(threshold)==1 && threshold<0 % dark particles
        pos=pos(~br1,:); % keep only dark particles
        br1=br1(~br1);
    end
    N=numel(br1);
    if ii==1 % if first frame, pre-allocate arrays for speed
        x=NaN(N*Nf,1);
        y=NaN(N*Nf,1);
        t=NaN(N*Nf,1);
        br=true(N*Nf,1);
        if ~all(arealim==[1 inf])
            ang=NaN(N*Nf,1);
        else
            ang=[]; % no angles
        end
        memloc=1;
    end
    if N>0
        x(memloc:memloc+N-1)=pos(:,1);
        y(memloc:memloc+N-1)=pos(:,2);
        t(memloc:memloc+N-1)=t1;
        br(memloc:memloc+N-1)=br1;
        if ~all(arealim==[1 inf])
            ang(memloc:memloc+N-1)=ang1;
        end
        memloc=memloc+N;
    end
    begins(ii+1)=begins(ii)+N;
    switch movtype
        case('seq')
            disp(['Found ' num2str(N,'%.0f') ' particles in ' ...
                names.name ' frame ' num2str(t1) ' ('  ...
                num2str(ii) ' of ' num2str(Nf) ').'])
        case('raw')
            disp(['Found ' num2str(N,'%.0f') ' particles in ' ...
                names.name ' frame ' num2str(t1) ' ('  ...
                num2str(ii) ' of ' num2str(Nf) ').'])
        case('avi')
            disp(['Found ' num2str(N,'%.0f') ' particles in ' ...
                names.name ' frame ' num2str(t1) ' ('  ...
                num2str(ii) ' of ' num2str(Nf) ').'])
        case('stack')
            disp(['Found ' num2str(N,'%.0f') ' particles in ' ...
                names.name ' frame ' num2str(t1) ' (' ...
                num2str(ii) ' of ' num2str(Nf) ').'])
        case('images')
            disp(['Found ' num2str(N,'%.0f') ' particles in ' ...
                names(ii).name ' (' num2str(ii) ' of ' num2str(Nf) ').'])
        case('array')
        disp(['Found ' num2str(N,'%.0f') ' particles in ' ...
             '(' num2str(ii) ' of ' num2str(Nf) ').'])
    end % switch movtype
    if writefile
        fwrite(fid,N,'int32'); % start the frame w/ particle count
        fwrite(fid,pos','float32'); % then write particle locations x1 y1 x2 y2  ...
        % Next: write angle as well!
    end
end % for ii=1:Nf
x(memloc:end)=[];
y(memloc:end)=[];
t(memloc:end)=[];
br(memloc:end)=[];
% Ensure all output is a column vector
x=x(:);
y=y(:);
t=t(:);
br=br(:);
if ~all(arealim==[1 inf])
    ang(memloc:end)=[];
    ang=ang(:);
end
if writefile
    fclose(fid);
end

% -=- Plot if requested -=------------------------------------------------
if noisy

    % -=- Arrange (x,y) coordinates for plotting -=-
    allt = min(t):max(t);
    Nt = numel(allt);
    [tt,begins] = unique(t);
    ends = circshift(begins-1,-1);
    Np = ends-begins+1;
    xy = NaN(max(Np),2,Nt);
    for ii=1:Nt
        indt = tt==allt(ii); % in case any frames have no particles
        ind = begins(indt):ends(indt);
        xy(1:Np(ii),:,ii) = [x(ind) y(ind)];
    end % for ii=1:Nt

    % -=- Read whole movie -=-
    switch movtype
        case('seq')
            im = read_seq(fullfile(filepath,names.name),[tmin tmax],0);
        case('raw')
            im = read_raw(fullfile(filepath,names.name),tmin:tmax,[],0);
            if numel(im)==2
                im{3} = 0*im{2}; % make imagei treat images as RGB
            end
        case('avi')
            vobj=VideoReader(fullfile(filepath,names.name));
%             for ii=1:tmin
%                 im = readFrame(vobj); % skip to first desired frame
%             end
            im =read(vobj,tmin);
            if strcmp(vobj.VideoFormat(1:3),'RGB')
                im = { repmat(im(:,:,1),[1 1 Nt]) ...
                    repmat(im(:,:,2),[1 1 Nt]) ...
                    repmat(im(:,:,3),[1 1 Nt]) };
                for ii=tmin+1:tmax
                    %im1 = readFrame(vobj);
                    im1 = read(vobj,ii);
                    im{1}(:,:,ii-tmin+1) = im1(:,:,1);
                    im{2}(:,:,ii-tmin+1) = im1(:,:,2);
                    im{3}(:,:,ii-tmin+1) = im1(:,:,3);
                end
            else % assume grayscale
                im = repmat(im,[1 1 Nt]);
                for ii=tmin+1:tmax
                    im(:,:,ii-tmin+1) = read(vobj,ii);
                end
            end % if strcmp(vobj.VideoFormat(1:3),'RGB')
        case('stack')
            im = imread(fullfile(filepath,names.name),tmin); 
            im = repmat(im,[1 1 Nt]);
            for ii=tmin+1:tmax
                im(:,:,ii-tmin+1) = ...
                    imread(fullfile(filepath,names.name),ii); 
            end
        case('images')
            im = imread(fullfile(filepath,names(tmin).name)); 
            im = repmat(im,[1 1 Nt]);
            for ii=tmin+1:tmax
                im(:,:,ii-tmin+1) = ...
                    imread(fullfile(filepath,names(ii).name)); 
            end
    end % switch movtype

    % -=- Let imagei do the rest -=-
    if ~isnumeric(noisy)
        imagei(im,[],xy,noisy)
    else
        imagei(im,[],xy)
    end

end % if noisy


end % function ParticleFinder

% -=- function FindParticles -=-------------------------------------------
function [pos,br] = FindParticles(im, threshold, logs)
% Given a 2D array "im", FindParticles finds small particles that are
% brighter than their four nearest neighbors and also brighter than
% "threshold", particles that are darker than their four nearest neighbors
% and also darker than "threshold", or both sets of particles. Particles
% are located to sub-pixel accuracy by applying a Gaussian fit in each
% spatial direction. The input "logs" depends on the color depth and is
% re-used for speed. Particle locations are returned in the two-column
% array "pos" (with x-coordinates in the first column and y-coordinates in
% the second). Particle contrast is indicated in the array "br": true if
% bright and false if dark. 

    s = size(im);

    if any(threshold>0) % find bright particles
        maxes = find(im >= threshold(threshold>0) & ...
            im > circshift(im,[0 1]) & ...
            im > circshift(im,[0 -1]) & ...
            im > circshift(im,[1 0]) & ...
            im > circshift(im,[-1 0]));
    else
        maxes = [];
    end

    if any(threshold<0) % find dark particles
        mins = find(im <= threshold(threshold<0) & ...
            im < circshift(im,[0 1]) & ...
            im < circshift(im,[0 -1]) & ...
            im < circshift(im,[1 0]) & ...
            im < circshift(im,[-1 0]));
    else
        mins = [];
    end

    br = [true(size(maxes)) ; false(size(mins))];

    % now turn these into subscripts
    [x,y] = ind2sub(s, [maxes ; mins]);

    % throw out unreliable locations in the outer ring
    good = find(x~=1 & y~=1 & x~=s(1) & y~=s(2));
    x = x(good);
    y = y(good);
    br = br(good);

    % find the horizontal positions

    % look up the logarithms of the relevant image intensities
    z1 = logs(abs(im(sub2ind(s,x-1,y))) + 1)';
    z2 = logs(abs(im(sub2ind(s,x,y))) + 1)';
    z3 = logs(abs(im(sub2ind(s,x+1,y))) + 1)';

    % compute the centers
    xcenters = -0.5 * (z1.*(-2*x-1) + z2.*(4*x) + z3.*(-2*x+1)) ./ ...
        (z1 + z3 - 2*z2);

    % do the same for the vertical position
    z1 = logs(abs(im(sub2ind(s,x,y-1))) + 1)';
    z3 = logs(abs(im(sub2ind(s,x,y+1))) + 1)';
    ycenters = -0.5 * (z1.*(-2*y-1) + z2.*(4*y) + z3.*(-2*y+1)) ./ ...
        (z1 + z3 - 2*z2);

    % make sure we have no bad points
    good = find(isfinite(xcenters) & isfinite(ycenters));

    % fix up the funny coordinate system used by matlab
    pos = [ycenters(good), xcenters(good)];
    br = br(good);

end % function FindParticles

% -=- function FindRegions -=---------------------------------------------
function [pos,ang,br] = FindRegions(im,threshold,arealim)
% Given a 2D array "im" and a two-element vector "arealim", FindRegions
% finds regions where the value of "im" exceeds "thresold" which also have
% area larger than arealim(1) and smaller than arealim(2) (in square
% pixels). If arealim is a scalar, it is taken as the minimum size. Region
% centroids are returned in the two-column array "pos" (with x-coordinates
% in the first column and y-coordinates in the second). Region orientations
% are returned in radians, in the vector "ang". Region contrast is
% indicated in the array "br": true if bright and false if dark. 

    if numel(arealim)==1
        arealim=[arealim inf]; % assume single size is a minimum
    end

    s = size(im);
    if any(threshold>0) % find bright particles
        propsBright=regionprops(im>=threshold(threshold>0),im, ...
            'WeightedCentroid','Area','PixelValues','PixelList');
    else
        propsBright = [];
    end
    if any(threshold<0) % find dark particles
        propsDark=regionprops(im<=threshold(threshold<0),im, ...
            'WeightedCentroid','Area','PixelValues','PixelList');
    else
        propsDark = [];
    end
    props = [ propsBright ; propsDark ];
    br = [ true(size(propsBright)) ; false(size(propsDark)) ];
    pos=vertcat(props.WeightedCentroid);
    if numel(pos)>0
        good = pos(:,1)~=1 & pos(:,2)~=1 & pos(:,1)~=s(2) & ...
            pos(:,2)~=s(1) & vertcat(props.Area)>arealim(1) & ...
            vertcat(props.Area)<arealim(2); % remove regions on edge, too small, or too big
        pos=pos(good,:);
        props=props(good);
        br=br(good);
    end
    ang = NaN(size(props));
    for ii = 1:numel(ang)
        xy = props(ii).PixelList;
        vals = props(ii).PixelValues;
        uxx = sum( (xy(:,1)-pos(ii,1)).^2 .* vals ,1) / sum(vals) + 1/12;
        uyy = sum( (xy(:,2)-pos(ii,2)).^2 .* vals ,1) / sum(vals) + 1/12;
        uxy = sum( (xy(:,1)-pos(ii,1)).*(xy(:,2)-pos(ii,2)) .* vals ,1) ...
            / sum(vals);
        if (uyy > uxx)
            num = uyy - uxx + sqrt((uyy - uxx)^2 + 4*uxy^2);
            den = 2*uxy;
        else
            num = 2*uxy;
            den = uxx - uyy + sqrt((uxx - uyy)^2 + 4*uxy^2);
        end
        if (num == 0) && (den == 0)
            ang(ii) = 0;
        else
            ang(ii) = atan(num/den);
        end
    end % for ii = 1:numel(ang)
end % function FindRegions
