classdef PredictiveTrackerClass
    %Class containing PredictiveTracks as a method and its inputs as
    %properties. The properties are saved as a 1x1 class, which should be
    %saved along with vtracks to maintain a record of the inputs used when
    %analyzing a video of particle motion. A good example video is located
    %in
    %/media/cage/glymphatic/20180912_examples/PTV/M22
    %
    %Example: to run predictivetrackerclass, type in command window:
    %obj = PredictiveTrackerClass;
    %obj.Thr= 3000;
    %obj.Arealim=5;
    %obj.Max_disp=10;
    %
    %To check and plot a few frames
    %obj.Bg=20; %using numeric background, value must be less than framerange
    %obj.Framerange = [1 50]
    %[vtracks,Ntr,Lmean,Lrms] = PredictiveTracks(obj,1)
    %
    %Change properties as desired using obj.____ syntax
    %Once satisfied, run
    %[vtracks,Ntr,Lmean,Lrms] = PredictiveTracks(obj,0)
    %or
    %[vtracks,Ntr,Lmean,Lrms] = PredictiveTracks(obj)
    %The properties can be checked by typing obj in the command line once
    %the object has been created. The class defaults to a raw file called
    %Image_0001_0001.raw and tracking over all frames
    %Be very careful chaning the defaults for filterwidth, n_predict and
    %mem. There is further information about those defaults in the comment
    %block of the properties

properties(Access = public)
    %These are the inputs needed for PredictiveTracks
    Fnames = 'Image_0001_0001.raw';
    Thr = NaN; 
    Framerange = [1 inf]; 
    Bg = 300;
    Arealim = NaN;
    Max_disp = NaN;
    Filterwidth=1;
    Chan=2;
    Mem=0; 
    N_predict = 1;
    %Notes on new functionality:
    %Mem: used to create a fake particles at the end of a track, which
    %can connect to a new track started later. For Mem=1, fake particles
    %are added for 1 frame, for Mem=2, fake particles are added for 2
    %frames etc. Useful for videos where particles are moving slightly in
    %and out of frame
    %N_predict: used to predict position of particle in next frame.
    %N_predict = 1 predicts based on velocity, N_predict = 2 predicts based 
    %on acceleration, 0 is no prediction. Maximum N_predic is 3
    %Filterwidth: determines the fit for velocity calculations, with
    %fitwidth always set to 3*filterwidth. Filterwidth=0 causes the
    %function to caluclate velocities using a central difference scheme
    %Masked_ROI: If a file named masked_ROI.mat exists in the workspace,
    %predictivetracks will only track particles in a specified region.
    
    
end
methods
%PredictiveTracker function
function [vtracks,Ntr,Lmean,Lrms] = PredictiveTracks(obj,noisy)
    % Usage: [vtracks,Ntr,Lmean,Lrms] =
    % PredictiveTracker(fnames,thr,max_disp,[framerange],[bg],[arealim],[noisy])
    %
    % Given a movie of particle motions, PredictiveTracker produces Lagrangian
    % particle tracks using a predictive three-frame best-estimate algorithm.
    % The movie must be saved in one of the following formats:
    % - a series of image files with sequential numerical filenames,
    % - an image stack in .tif or .gif format,
    % - a ThorLabs .raw movie with accompanying .xml file,
    % - a Norpix .seq movie, or
    % - a movie file readable by Matlab's VideoReader function.
    % Specify the movie in "fnames" (e.g., '0*.png' or 'stack.tif', or
    % 'movie.avi'). If thr<0, PredictiveTracker seeks particles darker than the
    % background by at least thr; if thr>0, PredictiveTracker seeks particles
    % brighter than the background by at least abs(thr); if thr is a
    % two-element array with thr(1)<0 and thr(2)>0, PredictiveTracker seeks
    % particles darker than the background by at least abs(thr(1)) or brighter
    % than the background by at least thr(2). If "bg" is the name of an image
    % file, PredictiveTracker reads the file and uses it as the background
    % image. If "bg" is an integer, PredictiveTracker calculates a dynamic
    % background using the mean of "bg" images surrounding the current frame.
    % If arealim==1, PredictiveTracker seeks single-pixel particles by
    % comparing brightness to adjacent pixels (fast and good for small
    % particles); otherwise PredictiveTracker seeks particles having areas
    % larger than arealim(1) and smaller than arealim(2) (both in square
    % pixels; this method is better for tracking large particles). Once
    % identified, each particle is tracked using a kinematic prediction, and a
    % track is broken when no particle lies within "max_disp" pixels of the
    % predicted location. The results are returned in the structure "vtracks",
    % whose fields "len", "X", "Y", "T", "U", and "V" contain the length,
    % horizontal coordinates, vertical coordinates, times, horizontal
    % velocities, and vertical velocities of each track, respectively. If
    % arealim~=1, "vtracks" is returned with an additonal field, "Theta",
    % giving the orientation of the major axis of the particle with respect to
    % the x-axis, in radians. If thr is a two-element array, "vtracks" is
    % returned an additional field, "isBright" which is true if the particle
    % was brighter than the background, and false otherwise. The total number
    % of tracks is returned as "Ntr"; the mean and root-mean-square track
    % lengths are returned in "Lmean" and "Lrms", respectively. If noisy==1, an
    % interactive visualization is produced, showing the original movie with
    % tracks overlaid. Requires ParticleFinder.m, and also requires
    % TracksPlotter.m for visualization. See also BackgroundImage.m. If
    % publications result from your use of this code, please cite "Using
    % particle tracking to measure flow instabilities in an undergraduate
    % laboratory experiment." D. H. Kelley and N. T. Ouellette. Am. J. Phys.
    % 79 (3) 267-273 (2011).

    % Written by Nicholas T. Ouellette September 2010.
    % Updated by Douglas H. Kelley 13 April 2011 to plot tracks.
    % Fixed bug in accounting active tracks 14 April 2011.
    % Allowed for fnames and/or bg in other directory, 4 May
    % 2011.
    % Fixed minor bug providing default input parameters 27 May 2011.
    % Added movie with overlaid velocity quivers 17 August 2011.
    % Added "finder" option 18 August 2011.
    % Changed "finder" to "minarea", added "invert" option, and combined two
    % plots into one 1 September 2011.
    % Enabled noisy>1 for saving a movie 2 September 2011.
    % Added "Theta" field (if minarea~=1) 7 September 2011.
    % Made compatible with uncompressed avi movies (using
    % read_uncompressed_avi.m) 19 October 2011.
    % Changed plotting method to use color and show only active tracks, 9
    % January 2012.
    % Improved compatibility with images having more than 255 grays, 10
    % January 2012.
    % Made compatible with tiff and gif stacks 13 February 2012.
    % Fixed bugs associated with frames in which no particles are found,
    % 6 March 2012.
    % Offloaded particle finding to ParticleFinder.m, 7 March 2012.
    % Updated 26 December 2013 to use VideoReader instead of
    % read_uncompressed_avi.m, and to be compatible with new version of
    % unique.m.
    % Updated 16 August 2014 for compatibility with Norpix .seq movies (using
    % Norpix2MATLAB.m).
    % Updated 4 June 2015 to use read_seq.m instead of Norpix2MATLAB.m, thereby
    % achieving compatibility with newer .seq files.
    % Updated 27 July 2015 to fix bug saving movies with single-file input.
    % Updated 24 January 2017 to fix bug saving movies in subdirectory.
    % Updated 31 October 2017 to properly handle frames with no particles.
    % Updated 29 March 2018, fixing another bug related to frames without
    % particles.
    % Updated 30 July 2018 to output isBright, to allow two-element threshold
    % input, and to drop "invert" input, since it's redundant.
    % Updated 1 August 2018 to allow dynamic backgrounds, using Jeff Tithof's
    % code. Also changed "noisy" input to allow a filename, and to save as .avi
    % instead of a directory full of images. Also renamed "minarea" to
    % "arealim".
    % Updated 11 July 2020 by Doug Kelley to be compatible with (new version
    % of?) readFrame.m: frame number is not an input.
    % Updated 9 March 2021 by Doug Kelley: Offloaded visualization to
    % TracksPlotter.m. Fixed bug where playback was wrong if framerange(1)>1.
    % Changed order of inputs to be like ParticleFinder.m.
    %Updated March 30th 2022 by Keelin Quirk to fit into a class as a method,
    %rather than as a stand alone function. Added fitwidth and filterwidth as
    %inputs.
    %April 6th 2022 - runs particlefinder, which has a channel input and
    %can run with an isolated region of interest called masked_ROI.mat
    %May 9th 2022- if filterwidth = 0, calculates velocity using a central 
    %difference scheme 
    %May 11th 2022 - has memory input to add up to Mem = n fake particles
    %when a track ends. 

    %-=- Initalize variables-=------------------------------------------------
    fnames = obj.Fnames;
    thr = obj.Thr;
    arealim = obj.Arealim;
    max_disp = obj.Max_disp;
    bg = obj.Bg;
    framerange=obj.Framerange;
    fitwidth=3*obj.Filterwidth;
    filterwidth=obj.Filterwidth;
    chan = obj.Chan;
    mem=obj.Mem;
    n_predict = obj.N_predict;
    %Set defaults
    noisy_default=0; % don't plot unless requested
    if ~exist('noisy','var') || isempty(noisy)
        noisy=noisy_default;
    end
    framerange_default=[1 inf]; % by default, all frames
    if ~exist('framerange','var') || isempty(framerange)
        framerange=framerange_default;
    end

    % -=- Find particles in all frames -=-------------------------------------
    [x,y,t,ang,br,allt]=Particlefinder(fnames,thr,framerange,[], ...
        bg,arealim,chan,0); % allt: list of all times
    Nf=numel(allt);
    if numel(x)==0
        error(['Sorry, found no particles in ' fnames '.'])
    end
    if verLessThan('matlab','8')
        [tt,ends]=unique(t); % tt: all times with particles
    else
        [tt,ends]=unique(t,'legacy');
    end
    begins=circshift(ends,[1 0])+1;
    begins(1)=1;
    
    % -=- Initialize tracks struct -=-----------------------------------------
    tracks = struct('len',[],'X',[],'Y',[],'T',[],'Theta',[],'fake',[],...
        'isBright',false);
    Nfake = 0;
    
    % -=- Loop over frames -=-------------------------------------------------
    for t=1:Nf
        indt = tt==allt(t); % necessary in case allt(1)~=1
        ind = begins(indt):ends(indt);
        Np = numel(ind)+Nfake; % particle count, depends on Nfake added in last frame
        %Stuff to keep track of fakes created each frame
        Nfake = 0;
        n=1;
        if Np>0
            if isempty(tracks(1).len)
                active = 1:Np;
                n_active = numel(active);
                for ii = 1:Np
                    tracks(ii).len=1;
                    tracks(ii).X=x(ind(ii));
                    tracks(ii).Y=y(ind(ii));
                    tracks(ii).T=allt(t);
                    tracks(ii).fake=0;
                end
                if arealim~=1
                    for ii = 1:Np
                        tracks(ii).Theta = ang(ind(ii));
                    end
                end
                if numel(thr)>1
                    for ii = 1:Np
                        tracks(ii).isBright = br(ind(ii));
                    end
                end
                tracks=tracks(:); % make it a column vector, not row
                continue
            end % if isempty(tracks(1).len)
            fr1=[x(ind) y(ind)];
            if any(arealim~=1)
                ang1=ang(ind);
            end
            if numel(thr)>1
                br1=br(ind);
            end
            %if fakes were created last frame, need to add those
            %particles to the arrays
            if exist('fakes','var')
                fr1=[fr1; fakes]; %append fake particles
                if any(arealim~=1)
                    angf = angf(:);
                    ang1=[ang1; angf]; %append fake particles
                end
                if numel(thr)>1
                    brf=brf(:);
                    br1=[br1; brf]; %append fake particles
                end
            end
            %clear fakes, etc. added previously
            clear fakes angf brf
            
            % -=- Match the tracks with kinematic predictions -=----------------------
            % estimate a velocity for each particle in fr0
            estimate = zeros(n_active,2);
            for ii = 1:n_active
                tr = tracks(active(ii));
                estimate(ii,1) = tr.X(end); % 0-frame prediction
                estimate(ii,2) = tr.Y(end);
                if n_predict > 0 && tr.len > 1
                    ux1 = tr.X(end) - tr.X(end-1);
                    uy1 = tr.Y(end) - tr.Y(end-1);
                    estimate(ii,1) = estimate(ii,1) + ux1; % 1-frame prediction (account for velocity)
                    estimate(ii,2) = estimate(ii,2) + uy1;
                    if n_predict > 1 && tr.len > 2
                        ux2 = tr.X(end-1) - tr.X(end-2);
                        uy2 = tr.Y(end-1) - tr.Y(end-2);
                        ax1 = ux1 - ux2;
                        ay1 = uy1 - uy2;
                        estimate(ii,1) = estimate(ii,1) + ax1/2; % 2-frame prediction (account for acceleration)
                        estimate(ii,2) = estimate(ii,2) + ay1/2;
                        if n_predict > 2 && tr.len > 3
                            ux3 = tr.X(end-2) - tr.X(end-3);
                            uy3 = tr.Y(end-2) - tr.Y(end-3);
                            ax2 = ux2 - ux3;
                            ay2 = uy2 - uy3;
                            estimate(ii,1) = estimate(ii,1) + (ax1-ax2)/6; % 3-frame prediction (account for jerk)
                            estimate(ii,2) = estimate(ii,2) + (ay1-ay2)/6;
                        end % if Npredict > 2 && tr.len > 3
                    end % if Npredict > 1 && tr.len > 2
                end % if Npredict > 0 && tr.len > 1
            end % for ii = 1:n_active
            
            % define cost and link arrays
            costs = zeros(n_active,1);
            links = zeros(n_active,1);
            
            % loop over active tracks
            for ii = 1:n_active
                % now, compare these estimated positions with particles in fr1
                dist_fr1 = (estimate(ii,1)-fr1(:,1)).^2 + (estimate(ii,2)-fr1(:,2)).^2;
                % save its cost and best match
                costs(ii) = min(dist_fr1);
                if costs(ii) > max_disp^2
                    continue;
                end
                bestmatch = find(dist_fr1 == costs(ii));
                % if there is more than one best match, we are confused; stop
                if numel(bestmatch) ~= 1
                    continue;
                end
                % has another track already matched to this particle?
                ind = links == bestmatch;
                if sum(ind) ~= 0
                    if costs(ind) > costs(ii) % this match is better
                        links(ind) = 0;
                    else % this match is not better
                        continue;
                    end
                end
                links(ii) = bestmatch;
            end % for ii = 1:n_active
            
            
            % now attach the matched particles to their tracks
            matched = zeros(Np,1);
            for ii = 1:n_active
                if links(ii) ~= 0
                    % this track found a match
                    tracks(active(ii)).X(end+1) = fr1(links(ii),1);
                    tracks(active(ii)).Y(end+1) = fr1(links(ii),2);
                    tracks(active(ii)).len = tracks(active(ii)).len + 1;
                    tracks(active(ii)).T(end+1) = allt(t);
                    tracks(active(ii)).fake(end+1)=0;
                    if any(arealim~=1)
                        tracks(active(ii)).Theta(end+1) = ang1(links(ii));
                    end
                    if numel(thr)>1
                        tracks(active(ii)).isBright(end+1) = br1(links(ii));
                    end
                    matched(links(ii)) = 1;
                    %If nothing is matched (links==0) , add a fake particle
                elseif mem>0
                    %see how many of the previous tracks were fake
                    num=0; %If there are none yet, need to have a number to start
                    for j=1:mem
                        if numel(tracks(active(ii)).fake)>=j
                            num(j)=tracks(active(ii)).fake(end+1-j);
                        end
                    end
                    numFake=sum(num);
                    %If there aren't too many fakes, create and mark a fake
                    %track
                    if numFake<mem
                        %keep track of fake particles
                        fakes(n,1) = estimate(ii,1);
                        fakes(n,2) = estimate(ii,2);
                        angf(n)=NaN;
                        brf(n)=NaN;
                        n=n+1;
                        %Add fake particle to track
                        tracks(active(ii)).X(end+1) = estimate(ii,1);
                        tracks(active(ii)).Y(end+1) = estimate(ii,2);
                        tracks(active(ii)).len = tracks(active(ii)).len + 1;
                        tracks(active(ii)).T(end+1) = allt(t);
                        tracks(active(ii)).fake(end+1)=1;
                        Nfake=Nfake+1; %also add a particle to next Np
                        if any(arealim~=1)
                            tracks(active(ii)).Theta(end+1) = NaN;
                        end
                        if numel(thr)>1
                            tracks(active(ii)).isBright(end+1) = NaN;
                        end
                        %Np=Np+1;
                        links(ii)=1;
                        matched(links(ii))=1;
                    end %if mem>0
                end % if links(ii) ~= 0
            end % for ii = 1:n_active
            active = active(links~=0);
            
            
            % and start new tracks with the particles in fr1 that found no match
            unmatched = find(matched == 0);
            newtracks = repmat(struct('len',[],'X',[],'Y',[],'T',[], ...
                'Theta',[],'fake',[],'isBright',[]),numel(unmatched),1);
            for ii = 1:numel(unmatched)
                newtracks(ii).len=1;
                newtracks(ii).X=fr1(unmatched(ii),1);
                newtracks(ii).Y=fr1(unmatched(ii),2);
                newtracks(ii).T=allt(t);
                newtracks(ii).fake=0;
            end
            if any(arealim~=1)
                for ii = 1:numel(unmatched)
                    newtracks(ii).Theta = ang1(unmatched(ii));
                end
            end
            if numel(thr)>1
                for ii = 1:numel(unmatched)
                    newtracks(ii).isBright = br1(unmatched(ii));
                end
            end
        else % if Np>0
            active = [];
            newtracks = [];
            unmatched = [];
            n_active = 0;
        end % if Np>0
        if exist('tracks','var')
            active = [active (numel(tracks)+1): ...
                (numel(tracks)+numel(newtracks))];
            tracks = [tracks ; newtracks];
            n_active = numel(active);
            links=[];
        end
        disp(['Frame ' num2str(t) ' of ' num2str(Nf) ': ' ...
            num2str(Np,'%.0f') ' particles, ' ...
            num2str(n_active,'%.0f') ' active, ' ...
            num2str(numel(unmatched),'%.0f') ' new, ' ...
            num2str(sum(links==0),'%.0f') ' unmatched, ' ...
            num2str(numel(tracks),'%.0f') ' total tracks.'])
        
    end % for t=1:Nf
    
    % -=- Prune tracks that are too short -=----------------------------------
    disp('Pruning...');
    %%need to add something that removes fake particles from the end
    
    if filterwidth >=1
        tracks = tracks([tracks.len] >= (2*fitwidth+1));
    else %filterwidth = 0 -> use CD scheme
        tracks = tracks([tracks.len] >= 3);
    end
    Ntr = numel(tracks);
    Lmean = mean([tracks.len]);
    Lrms = sqrt(mean([tracks.len].^2));
    
    % -=- Compute velocities -=-----------------------------------------------
    disp('Differentiating...');
    % loop over tracks
    vtracks = repmat(struct('len',[],'X',[],'Y',[],'T',[],'U',[],'V',[], ...
        'Theta',[],'fake',[],'isBright',false),Ntr,1);
    for ii = 1:Ntr
        %using CD in space if filterwidth =0
        if filterwidth ==0
            %Central difference for velocity, only drop first and last
            %particle from track
            for j=2:tracks(ii).len-1
                vtracks(ii).U(j-1)=(tracks(ii).X(j+1)-tracks(ii).X(j-1))/2;
                vtracks(ii).V(j-1)=(tracks(ii).Y(j+1)-tracks(ii).Y(j-1))/2;
            end
            vtracks(ii).len=tracks(ii).len - 2;
            vtracks(ii).X=tracks(ii).X(2:end-1);
            vtracks(ii).Y=tracks(ii).Y(2:end-1);
            vtracks(ii).T=tracks(ii).T(2:end-1);
            vtracks(ii).fake=tracks(ii).fake(2:end-1);
            if any(arealim~=1)
                vtracks(ii).Theta=tracks(ii).Theta(2:end-1);
            end
            if numel(thr)>1
                vtracks(ii).isBright = ...
                    tracks(ii).isBright(2:end-1);
            end
        else %gaussian approach using fitwidth and filterwidth to find velocity
            % define the convolution kernel
            Av = 1.0/(0.5*filterwidth^2 * ...
                (sqrt(pi)*filterwidth*erf(fitwidth/filterwidth) - ...
            2*fitwidth*exp(-fitwidth^2/filterwidth^2)));
        vkernel = -fitwidth:fitwidth;
        vkernel = Av.*vkernel.*exp(-vkernel.^2./filterwidth^2);
        u = -conv(tracks(ii).X,vkernel,'valid');
        v = -conv(tracks(ii).Y,vkernel,'valid');
        %create vtracks
        vtracks(ii).len=tracks(ii).len - 2*fitwidth;
        vtracks(ii).X=tracks(ii).X(fitwidth+1:end-fitwidth);
        vtracks(ii).Y=tracks(ii).Y(fitwidth+1:end-fitwidth);
        vtracks(ii).T=tracks(ii).T(fitwidth+1:end-fitwidth);
        vtracks(ii).U=u;
        vtracks(ii).V=v;
        vtracks(ii).fake=tracks(ii).fake(fitwidth+1:end-fitwidth);
        if any(arealim~=1)
            vtracks(ii).Theta=tracks(ii).Theta(fitwidth+1:end-fitwidth);
        end
        if numel(thr)>1
            vtracks(ii).isBright = ...
                tracks(ii).isBright(fitwidth+1:end-fitwidth);
        end
    end
end % for ii=1:Ntr

if all(arealim==1) % point-like particles, so don't output Theta
    vtracks=rmfield(vtracks,'Theta');
end
if numel(thr)==1 % all bright or all dark, so don't output isBright
    vtracks=rmfield(vtracks,'isBright');
end
disp('Done.')

% -=- Produce interactive plot if requested -=----------------------------
if ~isnumeric(noisy) || noisy==1
    TracksPlotter(vtracks,fnames,[],[],noisy)
end % if ~isnumeric(noisy) || noisy==1

    end % function PredictiveTracker
end
end