function hff = TracksPlotter(vtracks,fnames,fps,umperpix,MovName)
% Usage: hf = TracksPlotter(vtracks,[fnames],[fps],[umperpix],[MovName])
%
% Given particle tracks and the name of the movie from which they were
% produced, TracksPlotter produces an interactive visualization,
% reproducing the movie and overlaying the tracks. The user can change the
% time, play the movie (with tracks), change the brightness of the movie,
% change the coloring of the tracks, change whether the timer is displayed
% as a title or overlaid, and save a movie of the visualization to disk.
% Provide tracks in the struct array "vtracks", of the form produced by
% PredictiveTracker.m (or, to allow coloring according to downstream or
% cross-stream velocity, the form produced by calc_udotumeanhat.m).
% Optionally specify the movie in "fnames" (e.g., '0*.png' or 'stack.tif',
% or 'movie.avi'). If "fnames" is not given, tracks are plotted but the
% movie is not reproduced. Provide tracks with all times measured in
% frames, and all lengths in pixels. Optionally, specify the frame rate
% (e.g. in frames per second) in "fps" and/or the spatial resolution (e.g.
% in microns per pixel) in "umperpix", to visualize with units converted.
% The movie from which tracks were produced must be saved in one of the
% following formats: 
% - a series of image files with sequential numerical filenames,
% - an image stack in .tif or .gif format,
% - a ThorLabs .raw movie with accompanying .xml file,
% - a Norpix .seq movie, or
% - a movie file readable by Matlab's VideoReader function.
% 
% The graphics handle of the resulting figure is provided as "hf". To save
% a movie of the visualization, enter a file name ("movie name"; the
% optional input "MovName" is used by default) and click "save...". The
% duration, downsampling, and frame rate can then be adjusted. Requires
% superSlider.m. Also requires read_seq.m for .seq movies and read_raw.m
% for .raw movies.

% Written 10 March 2021 by Doug Kelley. 
% Updated 16 March 2021 by Doug Kelley to color-code tracks by
% instantaneous speed (not mean) and to improve layout.
% Updated 19 March 2021 to handle resizing window with arbitrary units. 
% 24 May 2021: Fixed bugs in plotting images. Added sliders for adjusting
% background movie brightness and contrast. 
% 25 May 2021: No longer drawing brightness slider if user provides no
% movie. No longer showing sliders when saving a movie.
% Update 1 June 2021: Default movie playback frame rate is now "fps", if
% it's provided, so playback is in real time. 
% Bug fixes 4 June 2021 to make axes fit in figure window. 
% Bug fix 2 July 2021: tab order of user controls is now more logical.
% Bug fix 6 July 2021: Sliders for adjusting background movie brightness
% and contrast now use units matching image file (not 0 to 1).
% 6 November 2021: Added dialog box to control timing, downsampling, and
% frame rate when saving. 
% 18 November 2021: Can now select file type when saving a movie. 
% 23 February 2022: Fixed bug in UpdateImageAndTracks that was causing
% incorrect image updates; now using read.m instead of readFrame.m!
% 26 May 2022: dhk added new color-coding options.
% 2 June 2022: Now runs much faster because dhk fixed a plotting bug (was
% updating the patch object repeatedly, in a loop). Also now works
% correctly when another TracksPlotter figure is already onscreen. Now
% uses frame rate encoded in the movie file, if one is available and "fps"
% isn't given. 
% 29 June 2022: No longer requires range.m. 

% Next: Add a "length unit" input and update colorbar label accordingly?

% -=- Constants -=---------------------------------------------------------
TrackLineWidth = 2; % for plotting
BackgroundColor = 0.94*[1 1 1];
FontSize = 8; % points
FigSize=[1024 768]; % figure width and height, in pixels
AxesBottomPos = 30; % bottom of axes, if multiple slices, pixels
AxesLeftPosSingle = 53; % left of axes, if one color slider
AxesLeftPosRGB = 156; % left of axes, if three color sliders
TimeLabelPos = [5 5 30 20]; % position of time label, pixels
TimeTextPos = [40 5 55 25]; % position of time text box, pixels
SliderPos = [105 7 130 20]; % position of time slider, pixels
OverlayButtonPos = [250 10 15 15]; % position of overlay checkbox, pixels
OverlayLabelPos = [270 5 55 27]; % position of overlay label, pixels
TimeUnitLabelPos = [330 5 40 27]; % position of time unit label, pixels
TimeUnitTextPos = [375 5 50 25]; % position of time unit text box, pixels
MovNameLabelPos = [435 5 50 27]; % position of movie name label, pixels
MovNameTextPos = [485 5 175 25]; % position of movie name text box, pixels
FrameRateLabelPos = [668 5 32 27]; % position of frame rate label, pixels
FrameRateTextPos = [705 5 50 25]; % position of frame rate text box, pixels
AxesButtonPos = [765 10 15 15]; % position of axes checkbox, pixels
AxesLabelPos = [785 5 35 27]; % position of axes label, pixels
ColorMenuPos = [825 5 140 25]; % position of color menu, pixels
PlayPauseButtonPos = [970 5 60 25]; % position of play button, pixels
Ch1Slider1ValuePos = [2 39 50 25];
Ch2Slider1ValuePos = [54 39 50 25];
Ch3Slider1ValuePos = [106 39 50 25];
Ch1SliderPos = [17 66 20 125];
Ch2SliderPos = [69 66 20 125];
Ch3SliderPos = [121 66 20 125];
Ch1Slider2ValuePos = [2 193 50 25];
Ch2Slider2ValuePos = [54 193 50 25];
Ch3Slider2ValuePos = [106 193 50 25];
Ch1LabelPos = [2 218 50 30];
Ch2LabelPos = [54 218 50 30];
Ch3LabelPos = [106 218 50 30];

MovieDialogSize = [400 220];
MoviePromptPos = [10 170 380 40];
MovieSliderPos = [10 140 380 20];
MovieStartTimeLabelPos = [10 110 90 20];
MovieStartTimeValuePos = [102 110 50 20];
MovieEndTimeLabelPos = [248 110 90 20];
MovieEndTimeValuePos = [340 110 50 20];
DownsampleLabelPos = [10 80 160 20];
DownsampleValuePos = [172 80 40 20];
NewFrameRateLabelPos = [215 80 103 20];
NewFrameRateValuePos = [320 80 70 20];
FileTypeLabelPos = [10 50 150 20];
FileTypeMenuPos = [162 50 228 20];
DurationPos = [10 10 220 20];
MovieCancelButtonPos = [260 10 60 30];
MovieOKButtonPos = [330 10 60 30];

% -=- Set defaults -=-----------------------------------------------------
FrameRate_default = 20; % frames per second, for playback only
DownSample = 1; % downsample factor, for saving
FileType_default = 'Motion JPEG AVI'; % for saving a movie

% -=- Parse inputs -=-----------------------------------------------------
assert(nargin>=1,['Usage: hf = ' ...
    mfilename '(vtracks,[fnames],[fps],[umperpix],[MovName])'])
if ~exist('fps','var') || isempty(fps)
    fps = 1;
    has_fps = false;
else
    FrameRate_default = fps;
    has_fps = true;
end
if ~exist('umperpix','var') || isempty(umperpix)
    umperpix = 1;
end
if ~exist('MovName','var') || isnumeric(MovName)
    MovName = [];
end

% -=- Plot everything for the first time -=-
defaultpos=get(0,'DefaultFigurePosition');
hf = figure('units','pixels','position',[defaultpos(1:2) FigSize]);
set(hf,'SizeChangedFcn',@SizeChanged); % combining with previous line causes an error, strangely
AxesLeftPos = 0;
ha = axes('nextplot','add','dataaspectratio',[1 1 1],'ydir','reverse', ...
    'box','on','units','pixels','outerposition', ...
    [AxesLeftPos AxesBottomPos FigSize(1)-AxesLeftPos ...
    FigSize(2)-AxesBottomPos],'parent',hf);
fr = min([vtracks.T]):max([vtracks.T]); % list of frames
t = fr/fps; % frame times
Nt = numel(t);
indt = 1; % first frame
t1 = t(indt);
ht=title(['t = ' num2str(t1,'%0.2f')]);
ht(2)=text(0,0,['t = ' num2str(t1,'%0.2f')]);
set(ht(2),'units','points','position',[5 5 0], ...
    'verticalalignment','bottom','horizontalalignment','left', ...
    'fontsize',FontSize,'fontweight','bold','backgroundcolor','k', ...
    'color','w','visible','off')
if ~exist('fnames','var') || isempty(fnames)
    movType = 'none';
    set(ha,'xlim',[min([vtracks.X]) max([vtracks.X])])
    set(ha,'ylim',[min([vtracks.Y]) max([vtracks.Y])])
else
    [filepath,~,ext]=fileparts(fnames);
    names=dir(fnames);
    if numel(names)==0
        error(['Sorry, cannot find ' fnames '.']);
    end
    if ( strcmpi(ext,'.avi') || ...
        strcmpi(ext,'.mp4') || strcmpi(ext,'.mov') || strcmpi(ext,'.mj2') || ...
        strcmpi(ext,'.wmv') || strcmpi(ext,'.mpg') || strcmpi(ext,'.asf') || ...
        strcmpi(ext,'.asx') || strcmpi(ext,'.m4v') || strcmpi(ext,'.ogg') ) % single movie file
        movType = 'VideoReader';
        vobj=VideoReader(fullfile(filepath,names.name));
        if ~has_fps
            FrameRate_default = vobj.FrameRate;
        end
        im = read(vobj,fr(indt));
    elseif strcmpi(ext,'.seq')
        movType = 'seq';
        [im,~,info] = read_seq(fnames,fr(indt),0);
        if ~has_fps
            FrameRate_default = info.FrameRate;
        end
    elseif strcmpi(ext,'.raw')
        movType = 'raw';
        [im1,~,info] = read_raw(fnames,fr(indt),[],0);
        if ~has_fps
            FrameRate_default = info.FrameRate;
        end
        Nch=numel(im1);
        if Nch==1
            im=im1{1};
        else
            im=repmat(0*im1{1}(:,:,1),[1 1 3]); % initialize
            for ii=1:Nch
                im(:,:,ii)=im1{ii};
            end
        end
    elseif numel(names)==1 && ( strcmpi(ext,'.tif') || ...
            strcmpi(ext,'.tiff') || strcmpi(ext,'.gif') ) % single file, looks like an image stack
        movType = 'stack';
        im = imread(fullfile(filepath,names.name),fr(indt));
    else
        movType = 'images';
        im = imread(fullfile(filepath,names(fr(indt)).name));
    end % if strcmpi(ext,'.avi') || ...
    if ismatrix(im)
        isRGB = false;
        AxesLeftPos = AxesLeftPosSingle;
        im = repmat(im,[1 1 3]); % force to grayscale
    else
        isRGB = true;
        AxesLeftPos = AxesLeftPosRGB;
    end
    set(ha,'outerposition', ...
        [AxesLeftPos AxesBottomPos FigSize(1)-AxesLeftPos ...
        FigSize(2)-AxesBottomPos]);
    ch1m = double(min(min(im(:,:,1))));
    ch2m = double(min(min(im(:,:,2))));
    ch3m = double(min(min(im(:,:,3))));
    ch1r = double(max(max(im(:,:,1))))-ch1m;
    ch2r = double(max(max(im(:,:,2))))-ch2m;
    ch3r = double(max(max(im(:,:,3))))-ch3m;
    if ch1r==0  % avoid NaNs later
        ch1r=1;
    end
    if ch2r==0  % avoid NaNs later
        ch2r=1;
    end
    if ch3r==0  % avoid NaNs later
        ch3r=1;
    end
    clim_ch1 = ch1m + [0 ch1r];
    clim_ch2 = ch2m + [0 ch2r];
    clim_ch3 = ch3m + [0 ch3r];
    sz = size(im);
    imscaled = double(im);
    imscaled(:,:,1) = (imscaled(:,:,1)-clim_ch1(1)) / diff(clim_ch1);
    imscaled(:,:,2) = (imscaled(:,:,2)-clim_ch2(1)) / diff(clim_ch2);
    imscaled(:,:,3) = (imscaled(:,:,3)-clim_ch3(1)) / diff(clim_ch3);
    imscaled(imscaled<0) = 0;
    hi=imagesc((0:sz(2))*umperpix,(0:sz(1))*umperpix,imscaled, ...
        'parent',ha);
    set(ha,'xlim',(0.5+[0 size(im,2)])*umperpix);
    set(ha,'ylim',(0.5+[0 size(im,1)])*umperpix);
end % if ~exist('fnames','var') || isempty(fnames)
Ntr = numel(vtracks);
tr_fr_range = NaN(Ntr,2); % range of active frames, for each track
for ii=1:Ntr
    tr_fr_range(ii,:) = [vtracks(ii).T(1) vtracks(ii).T(end)];
end
ind = find( (tr_fr_range(:,1)<=fr(1)) ...
    & (tr_fr_range(:,2)>=fr(1)) );
Np = numel(ind); % count of particle tracks active in current frame
X1 = NaN(2*Np,1);
Y1 = NaN(2*Np,1);
C1 = NaN(2*Np,1);
for ii=1:Np
    X1(2*ii-1) = vtracks(ind(ii)).X(1)*umperpix;
    Y1(2*ii-1) = vtracks(ind(ii)).Y(1)*umperpix;
    C1(2*ii-1) = sqrt( vtracks(ind(ii)).U(1).^2 ...
        + vtracks(ind(ii)).V(1).^2 )*umperpix*fps;
end
hp = patch(X1,Y1,C1,'edgecolor','interp','linewidth',TrackLineWidth, ...
    'parent',ha);
spd = sqrt( [vtracks.U].^2 + [vtracks.V].^2 )*umperpix*fps;
caxis([min(spd) max(spd)]) % must set because of apparent bug in Matlab: colorbar doesn't update when patch objects change color
hcb = colorbar('parent',hf);
hcb.Label.String = 'instantaneous speed'; % assumed to be the default
set(hf,'name',['t = ' num2str(t(indt),'%0.2f') ' : ' num2str(Np) ...
    ' particles'])

% -=- Make controls -=-
hr = annotation('rectangle',[0 0 1 1]);
set(hr,'color','r','units','pixels','position',[AxesLeftPos ...
    AxesBottomPos FigSize(1)-AxesLeftPos FigSize(2)-AxesBottomPos], ...
    'linewidth',5,'facecolor','k','facealpha',0.2,'visible','off');
uicontrol('Style','text','units','pixels', ... % Time label
    'Position',TimeLabelPos,'FontSize',FontSize,'FontWeight','bold', ...
    'String','t = ','ForegroundColor','k','parent',hf, ...
    'horizontalalignment','right','backgroundcolor',BackgroundColor);
TimeText = uicontrol('Style','edit','units','pixels', ...
    'Position',TimeTextPos,'String',num2str(t(1),'%.5g'), ...
    'fontsize',FontSize,'Callback',@inputFromTimeText, ...
    'parent',hf);
TimeSlider = uicontrol('Style','slider','units','pixels', ...
    'Position',SliderPos,'Min',t(1),'Max',t(end), ...
    'Value',t(1),'SliderStep', ...
    [1/(max(fr)-min(fr)) 1/(max(fr)-min(fr))+eps], ...
    'Callback',@inputFromTimeSlider,'parent',hf);
OverlayButton = uicontrol('style','checkbox', ...
    'units','pixels','position',OverlayButtonPos, ...
    'string',[],'value',0,'parent',hf, ...
    'backgroundcolor',BackgroundColor,'callback',@OverlayToggle);
uicontrol('Style','text','units','pixels', ... % overlay label
    'Position',OverlayLabelPos,'FontSize',FontSize,'FontWeight','bold', ...
    'String','overlay timer','ForegroundColor','k', ...
    'horizontalalignment','left','backgroundcolor',BackgroundColor, ...
    'parent',hf);
uicontrol('Style','text','units','pixels', ... % Time unit label
    'Position',TimeUnitLabelPos,'FontSize',FontSize,'parent',hf, ...
    'FontWeight','bold','String','time unit:','ForegroundColor','k', ...
    'horizontalalignment','right','backgroundcolor',BackgroundColor);
TimeUnitText = uicontrol('Style','edit','units','pixels', ...
    'Position',TimeUnitTextPos,'String','','parent',hf, ...
    'fontsize',FontSize,'Callback',@inputFromTimeUnitText);
uicontrol('Style','text','units','pixels', ... % Movie name label
    'Position',MovNameLabelPos,'FontSize',FontSize, ...
    'FontWeight','bold','String','movie name: ', ...
    'ForegroundColor','k','horizontalalignment','right', ...
    'backgroundcolor',BackgroundColor,'parent',hf);
MovNameText = uicontrol('Style','edit','units','pixels', ...
    'fontsize',FontSize,'Position',MovNameTextPos,'String',MovName, ...
    'callback',@inputFromMovNameText,'parent',hf);
uicontrol('Style','text','units','pixels', ... % Frame rate label
    'Position',FrameRateLabelPos,'FontSize',FontSize,'parent',hf, ...
    'FontWeight','bold','String','play fps: ','ForegroundColor','k', ...
    'horizontalalignment','right','backgroundcolor',BackgroundColor);
FrameRateText = uicontrol('Style','edit','units','pixels', ...
    'fontsize',FontSize,'Position',FrameRateTextPos, ...
    'String',FrameRate_default,'parent',hf);
AxesButton = uicontrol('style','checkbox','parent',hf, ...
    'units','pixels','position',AxesButtonPos,'string',[], ...
    'value',0,'backgroundcolor',BackgroundColor,'callback',@AxesToggle);
uicontrol('Style','text','units','pixels', ... % axes button label
    'Position',AxesLabelPos,'FontSize',FontSize,'FontWeight','bold', ...
    'String','axes only','ForegroundColor','k','parent',hf, ...
    'horizontalalignment','left','backgroundcolor',BackgroundColor);
ColorMenu = uicontrol('style','popupmenu','parent',hf, ...
    'units','pixels','position',ColorMenuPos,'fontsize',FontSize, ...
    'string',{'instantaneous speed','mean speed','time','none'}, ...
    'callback',@inputFromColorMenu);
if any(strcmp(fieldnames(vtracks),'Udown'))
    ColorMenu.String = [ColorMenu.String ; 'downstream velocity'];
end
if any(strcmp(fieldnames(vtracks),'Ucross'))
    ColorMenu.String = [ColorMenu.String ; 'cross-stream velocity'];
end
PlayPauseButton = uicontrol('style','pushbutton', ...
    'units','pixels','position',PlayPauseButtonPos, ...
    'string','play','callback',@PlayPause,'parent',hf);
if ~isempty(get(MovNameText,'string'))
    set(PlayPauseButton,'string','save...')
end
if ~strcmpi(movType,'none')
    Ch1Slider=superSlider(gcf,'position', ...
        Ch1SliderPos,'value', [0 1], 'numslides',2,'max',1,'min',0, ...
        'stepsize',.05,'Callback',@inputFromCh1Slider);
    set(Ch1Slider,'units','pixels')
    set(Ch1Slider,'position',Ch1SliderPos,'parent',hf)
    Ch1Label = uicontrol('Style','text','units','pixels', ... % Channel 1 label
        'Position',Ch1LabelPos,'FontSize',FontSize,'FontWeight','bold', ...
        'String','Ch 1','ForegroundColor','k','parent',hf, ...
        'horizontalalignment','center','backgroundcolor',BackgroundColor);
    Ch1Slider2Value = uicontrol('Style','edit','units','pixels', ...
        'Position',Ch1Slider2ValuePos,'String',num2str(clim_ch1(2)), ...
        'fontsize',FontSize,'Callback',@inputFromCh1Slider2Value, ...
        'parent',hf);
    if isRGB
        Ch2Slider=superSlider(gcf,'position', ...
            Ch2SliderPos,'value', [0 1], 'numslides',2,'max',1,'min',0, ...
            'stepsize',.05,'Callback',@inputFromCh2Slider);
        set(Ch2Slider,'units','pixels')
        set(Ch2Slider,'position',Ch2SliderPos,'parent',hf)
        uicontrol('Style','text','units','pixels', ... % Channel 2 label
            'Position',Ch2LabelPos,'FontSize',FontSize,'FontWeight','bold', ...
            'String','Ch 2 (green)','ForegroundColor','k','parent',hf, ...
            'horizontalalignment','center','backgroundcolor',BackgroundColor);
        Ch3Slider=superSlider(gcf,'position', ...
            Ch3SliderPos,'value', [0 1], 'numslides',2,'max',1,'min',0, ...
            'stepsize',.05,'Callback',@inputFromCh3Slider);
        set(Ch3Slider,'units','pixels')
        set(Ch3Slider,'position',Ch3SliderPos,'parent',hf)
        uicontrol('Style','text','units','pixels', ... % Channel 3 label
            'Position',Ch3LabelPos,'FontSize',FontSize,'FontWeight','bold', ...
            'String','Ch 3 (blue)','ForegroundColor','k','parent',hf, ...
            'horizontalalignment','center','backgroundcolor',BackgroundColor);
        Ch2Slider2Value = uicontrol('Style','edit','units','pixels', ...
            'Position',Ch2Slider2ValuePos,'String',num2str(clim_ch2(2)), ...
            'fontsize',FontSize,'Callback',@inputFromCh2Slider2Value, ...
            'parent',hf);
        Ch3Slider2Value = uicontrol('Style','edit','units','pixels', ...
            'Position',Ch3Slider2ValuePos,'String',num2str(clim_ch3(2)), ...
            'fontsize',FontSize,'Callback',@inputFromCh3Slider2Value, ...
            'parent',hf);
        set(Ch1Label,'string','Ch 1 (red)')
    end % if isRGB
    Ch1Slider1Value = uicontrol('Style','edit','units','pixels', ...
        'Position',Ch1Slider1ValuePos,'String',num2str(clim_ch1(1)), ...
        'fontsize',FontSize,'Callback',@inputFromCh1Slider1Value, ...
        'parent',hf); 
    if isRGB % the extra if-then is to get the tab order right
        Ch2Slider1Value = uicontrol('Style','edit','units','pixels', ...
            'Position',Ch2Slider1ValuePos,'String',num2str(clim_ch2(1)'), ...
            'fontsize',FontSize,'Callback',@inputFromCh2Slider1Value, ...
            'parent',hf);
        Ch3Slider1Value = uicontrol('Style','edit','units','pixels', ...
            'Position',Ch3Slider1ValuePos,'String',num2str(clim_ch3(1)), ...
            'fontsize',FontSize,'Callback',@inputFromCh3Slider1Value, ...
            'parent',hf);
    end % if isRGB
end % if ~strcmpi(movType,'none')
if nargout>0
    hff = hf;
end

% -=- Callback: When the user resizes the window -=-
function SizeChanged(~,~)
    old_ax_units = get(ha,'units');
    old_fig_units = get(hf,'units');
    set(hf,'units','pixels');
    hf_pos = get(hf,'position');
    set(ha,'units','pixels','outerposition', ...
        [AxesLeftPos AxesBottomPos hf_pos(3)-AxesLeftPos ...
        hf_pos(4)-AxesBottomPos]);
    set(ha,'units',old_ax_units);
    set(hf,'units',old_fig_units);
end % function SizeChanged(~,~)

% -=- Callback: When the user changes time unit -=-
    function inputFromTimeUnitText(~,~)
        t1=get(TimeSlider,'value');
        changeTime();
    end % function inputFromTimeUnitText(~,~)

% -=- Callback: When the user changes time using the slider bar -=-
    function inputFromTimeSlider(~,~)
        t1=get(TimeSlider,'value');
        changeTime();
    end % function inputFromTimeSlider(~,~)

% -=- Callback: When the user changes time using the text box -=-
    function inputFromTimeText(~,~)
        t1 = str2double(get(TimeText,'string'));
        if isnan(t1) % user entered a non-number; replace with current time
            set(TimeText,'string',num2str(t(indt),'%.5g'));
            return
        end
        changeTime();
    end % function inputFromTimeText(~,~)

% -=- Callback: When the user toggles axes checkbox -=-
    function AxesToggle(~,~)
        if get(AxesButton,'value')
            li = GetLayoutInformation(ha);
            set(hr,'position',li.PlotBox);
        else
            old_units = get(hf,'units');
            set(hf,'units','pixels');
            hf_pos = get(hf,'position');
            set(hr,'position',[AxesLeftPos AxesBottomPos ...
                hf_pos(3)-AxesLeftPos ...
                hf_pos(4)-AxesBottomPos]);
            set(hf,'units',old_units);
        end
        set(hr,'visible','on')
        pause(1)
        set(hr,'visible','off')
    end % function AxesToggle(~,~)

% -=- Callback: When the user toggles overlay checkbox -=-
    function OverlayToggle(~,~)
        if get(OverlayButton,'value')
            set(ht(1),'visible','off')
            set(ht(2),'visible','on')
        else
            set(ht(1),'visible','on')
            set(ht(2),'visible','off')
        end
    end % function OverlayToggle(~,~)

% -=- Callback: When user changes movie name -=-
    function inputFromMovNameText(~,~)
        if isempty(get(MovNameText,'string'))
            set(PlayPauseButton,'string','play')
        else
            set(PlayPauseButton,'string','save...')
        end
    end

% -=- Callback: When user clicks PlayPauseButton -=-
    function PlayPause(~,~)
        if strcmpi(get(PlayPauseButton,'string'),'play') || ...
            strcmpi(get(PlayPauseButton,'string'),'save...') % could implement "Save" as a separate function, but setup and cleanup duplicate "Play"
            Play()
        else
            Pause()
        end
    end % function PlayPause(~,~)

% -=- Callback: When the user changes Channel limit Sliders -=-
    function inputFromCh1Slider(~,~)
        ch1lim=get(Ch1Slider, 'UserData');
        clim_ch1=ch1lim(1,:)*ch1r+ch1m;    
        if ~isRGB
            clim_ch2 = clim_ch1;
            clim_ch3 = clim_ch1;
        end
        set(Ch1Slider1Value,'String', num2str(clim_ch1(1),'%0.5g'))
        set(Ch1Slider2Value,'String', num2str(clim_ch1(2),'%0.5g'))
        updateImageAndTracks();
    end 
    function inputFromCh2Slider(~,~)
        ch2lim=get(Ch2Slider, 'UserData');
        clim_ch2=ch2lim(1,:)*ch2r+ch2m;
        set(Ch2Slider1Value,'String', num2str(clim_ch2(1),'%0.5g'))
        set(Ch2Slider2Value,'String', num2str(clim_ch2(2),'%0.5g'))
        updateImageAndTracks();
    end 
    function inputFromCh3Slider(~,~)
        ch3lim=get(Ch3Slider, 'UserData');
        clim_ch3=ch3lim(1,:)*ch3r+ch3m;
        set(Ch3Slider1Value,'String', num2str(clim_ch3(1),'%0.5g'))
        set(Ch3Slider2Value,'String', num2str(clim_ch3(2),'%0.5g'))
        updateImageAndTracks();
    end 

% -=- Callback: When the user changes color menu -=-
    function inputFromColorMenu(~,~)
        updateImageAndTracks();
        caxis auto
    end

% -=- Callback: When the user changes Channel limit Value Boxes -=-
    function inputFromCh1Slider1Value(~,~)
        limtmp=str2double(get(Ch1Slider1Value,'string')); 
        if isnan(limtmp)
            set(Ch1Slider1Value,'string',num2str(clim_ch1(1),'%0.5g'))
            return
        end
        clim_ch1(1)=limtmp;
        if ~isRGB
            clim_ch2(1) = limtmp;
            clim_ch3(1) = limtmp;
        end
        updateImageAndTracks();
        
        % update slider location
        allSlides = get(Ch1Slider, 'Children');
        location = get(allSlides(1), 'Position');
        location(2)=((clim_ch1(1)-ch1m)/ch1r)*(1-2*location(4)); % scale since the maximum location is 1-2*location(4) 
        set(allSlides(1),'Position',location)        
    end 
    function inputFromCh2Slider1Value(~,~)
        limtmp=str2double(get(Ch2Slider1Value,'string')); 
        if isnan(limtmp)
            set(Ch2Slider1Value,'string',num2str(clim_ch2(1),'%0.5g'))
            return
        end
        clim_ch2(1)=limtmp;
        updateImageAndTracks();
        
        % update slider location
        allSlides = get(Ch2Slider, 'Children');
        location = get(allSlides(1), 'Position');
        location(2)=((clim_ch2(1)-ch2m)/ch2r)*(1-2*location(4)); % scale since the maximum location is 1-2*location(4) 
        set(allSlides(1),'Position',location) 
    end 
    function inputFromCh3Slider1Value(~,~)
        limtmp=str2double(get(Ch3Slider1Value,'string')); 
        if isnan(limtmp)
            set(Ch3Slider1Value,'string',num2str(clim_ch3(1),'%0.5g'))
            return
        end
        clim_ch3(1)=limtmp;
        updateImageAndTracks();
                        
        % update slider location
        allSlides = get(Ch3Slider, 'Children');
        location = get(allSlides(1), 'Position');
        location(2)=((clim_ch3(1)-ch3m)/ch3r)*(1-2*location(4)); % scale since the maximum location is 1-2*location(4) 
        location(2)=0.5;
        set(allSlides(1),'Position',location) 
    end 
    function inputFromCh1Slider2Value(~,~)
        limtmp=str2double(get(Ch1Slider2Value,'string')); 
        if isnan(limtmp)
            set(Ch1Slider2Value,'string',num2str(clim_ch1(2),'%0.5g'))
            return
        end
        clim_ch1(2)=limtmp;
        if ~isRGB
            clim_ch2(2) = limtmp;
            clim_ch3(2) = limtmp;
        end
        updateImageAndTracks();
        
        % update slider location
        allSlides = get(Ch1Slider, 'Children');
        location = get(allSlides(2), 'Position');
        location(2)=((clim_ch1(2)-ch1m)/ch1r)*(1-2*location(4))+location(4); % scale since the minimum location is location(4) and the max is 1-location(4)
        set(allSlides(2),'Position',location)
    end 
    function inputFromCh2Slider2Value(~,~)
        limtmp=str2double(get(Ch2Slider2Value,'string')); 
        if isnan(limtmp)
            set(Ch2Slider2Value,'string',num2str(clim_ch2(2),'%0.5g'))
            return
        end
        clim_ch2(2)=limtmp;
        updateImageAndTracks();
        
        % update slider location
        allSlides = get(Ch2Slider, 'Children');
        location = get(allSlides(2), 'Position');
        location(2)=((clim_ch2(2)-ch2m)/ch2r)*(1-2*location(4))+location(4); % scale since the minimum location is location(4) and the max is 1-location(4)
        set(allSlides(2),'Position',location)
    end 
    function inputFromCh3Slider2Value(~,~)
        limtmp=str2double(get(Ch3Slider2Value,'string')); 
        if isnan(limtmp)
            set(Ch3Slider2Value,'string',num2str(clim_ch3(2),'%0.5g'))
            return
        end
        clim_ch3(2)=limtmp;
        updateImageAndTracks();
        
        % update slider location
        allSlides = get(Ch3Slider, 'Children');
        location = get(allSlides(2), 'Position');
        location(2)=((clim_ch3(2)-ch3m)/ch3r)*(1-2*location(4))+location(4); % scale since the minimum location is location(4) and the max is 1-location(4)
        set(allSlides(2),'Position',location)
    end 

% -=- Start playback -=-
    function Play(~)

        set([TimeText MovNameText FrameRateText OverlayButton ...
            TimeUnitText],'enable','off'); % disable controls
        set(TimeSlider,'callback',''); % disable slider while movie plays
        FrameRate=str2double(get(FrameRateText,'string'));
        if isnan(FrameRate)
            FrameRate = FrameRate_default;
            set(FrameRateText,'string',num2str(FrameRate))
        end
        MovName=get(MovNameText,'string');
        if ~isempty(MovName) % save playback to disk
            set(PlayPauseButton,'string','pause','enable','off')
            indt0=indt; % save the time for later
            [~,~,ext]=fileparts(MovName);
            if ~strcmpi(ext,'.avi')
                MovName=[MovName '.avi']; % append extension if necessary
            end
            old_units = get(hf,'units');
            MovieTimes = [t(1) t(end)];

            % -=- Make dialog box and controls -=-
            figPos = get(gcf,'position');
            MovieDialog = dialog('position', ...
                [figPos(1)+figPos(3)/2-MovieDialogSize(1)/2 ...
                    figPos(2)+figPos(4)/2-MovieDialogSize(2)/2 ...
                    MovieDialogSize], ...
                'name',['Saving ' MovName ' ...']);
            uicontrol(MovieDialog,'style','text','units','pixels', ...
                'position',MoviePromptPos, ...
                'foregroundcolor','k', ...
                'backgroundcolor',BackgroundColor,'string', ...
                ['Please specify the times, downsampling, and frame ' ...
                'rate. The original frame rate was ' num2str(fps) ' Hz.'])
            MovieDialogSlider = superSlider(MovieDialog, ...
                'position',MovieSliderPos, ...
                'value',[t(1) t(end)],'numslides',2, ...
                'max',t(end),'min',t(1),'stepsize',(max(t)-min(t))/20, ...
                'callback',@inputFromMovieDialogSlider);
            set(MovieDialogSlider,'units','pixels')
            set(MovieDialogSlider,'position',MovieSliderPos)
            uicontrol(MovieDialog,'style','text','units','pixels', ...
                'position',MovieStartTimeLabelPos,'foregroundcolor','k', ...
                'backgroundcolor',BackgroundColor, ...
                'string','start time (s):', ...
                'horizontalalignment','right')
            MovieStartTimeValue = uicontrol(MovieDialog,'style','edit', ...
                'units','pixels','position',MovieStartTimeValuePos, ...
                'string',num2str(t(1),'%0.2f'), ...
                'callback',@inputFromMovieStartTimeValue);
            uicontrol(MovieDialog,'style','text','units','pixels', ...
                'position',MovieEndTimeLabelPos,'foregroundcolor','k', ...
                'backgroundcolor',BackgroundColor, ...
                'string','end time (s):', ...
                'horizontalalignment','right')
            MovieEndTimeValue = uicontrol(MovieDialog,'style','edit', ...
                'units','pixels','position',MovieEndTimeValuePos, ...
                'string',num2str(t(end),'%0.2f'), ...
                'callback',@inputFromMovieEndTimeValue);
            uicontrol(MovieDialog,'style','text','units','pixels', ...
                'position',DownsampleLabelPos,'foregroundcolor','k', ...
                'backgroundcolor',BackgroundColor, ...
                'string','downsample factor:', ...
                'horizontalalignment','right')
            DownsampleValue = uicontrol(MovieDialog,'style','edit', ...
                'units','pixels','position',DownsampleValuePos, ...
                'string',num2str(DownSample), ...
                'callback',@inputFromDownsampleValue);
            uicontrol(MovieDialog,'style','text','units','pixels', ...
                'position',NewFrameRateLabelPos,'foregroundcolor','k', ...
                'backgroundcolor',BackgroundColor, ...
                'string','frame rate (Hz):', ...
                'horizontalalignment','right')
            NewFrameRateValue = uicontrol(MovieDialog,'style','edit', ...
                'units','pixels','position',NewFrameRateValuePos, ...
                'string',num2str(FrameRate,'%0.3f'), ...
                'callback',@inputFromNewFrameRateValue);
            uicontrol(MovieDialog,'style','text','units','pixels', ...
                'position',FileTypeLabelPos,'foregroundcolor','k', ...
                'backgroundcolor',BackgroundColor, ...
                'string','file type:','horizontalalignment','right')
            if strcmp(computer,'GLNXA64') % Linux / Unix
                FileTypeList = {'Motion JPEG AVI','Archival'};
            else % Max / Windows
                FileTypeList = {'Motion JPEG AVI','MPEG-4','Archival'};
            end
            FileType = FileType_default;
            FileTypeMenu = uicontrol(MovieDialog,'style','popupmenu', ...
                'units','pixels','position', ...
                FileTypeMenuPos,'foregroundcolor','k', ...
                'backgroundcolor',BackgroundColor,'string', ...
                FileTypeList,'horizontalalignment','left', ...
                'callback',@inputFromFileTypeMenu);
            DurationMessage = uicontrol(MovieDialog,'style','text', ...
                'units','pixels','position', ...
                DurationPos,'foregroundcolor','k', ...
                'backgroundcolor',BackgroundColor,'string', ...
                ['new duration: ' num2str(max(t)-min(t),'%0.2f') ' s'], ...
                'horizontalalignment','left');
            uicontrol(MovieDialog,'style','pushbutton', ...
                'units','pixels','position',MovieCancelButtonPos, ...
                'string','Cancel','callback',@inputFromMovieCancelButton);
            uicontrol(MovieDialog,'style','pushbutton', ...
                'units','pixels','position',MovieOKButtonPos, ...
                'string','OK','callback',@inputFromMovieOKButton);
            waitfor(MovieDialog);

            set(hf,'units',old_units);
            t1=t(indt0);
            set([TimeText MovNameText FrameRateText OverlayButton ...
                TimeUnitText PlayPauseButton],'enable','on'); % enable controls
            if isempty(get(MovNameText,'string'))
                set(PlayPauseButton,'string','play')
            else
                set(PlayPauseButton,'string','save...')
            end
            set(TimeSlider,'callback',@inputFromTimeSlider); % enable slider
            FrameRate=str2double(get(FrameRateText,'string')); % restore old frame rate
            changeTime() % go back to the same time

        else % if ~isempty(MovName) - just play, don't save
            set(PlayPauseButton,'string','pause')
            while strcmpi(get(PlayPauseButton,'string'),'pause')
                indt=indt+1;
                if indt>Nt
                    indt=1; % loop it
                end
                t1=t(indt);
                changeTime();
                pause(1/FrameRate);
            end % while strcmpi(get(PlayPauseButton,'string'),'pause')
        end % if ~isempty(MovName)

% -=- Callback: When the user changes movie dialog slider -=-
    function inputFromMovieDialogSlider(~,~)
        limtmp = get(MovieDialogSlider,'userdata');
        MovieTimes = limtmp(1,:);
        set(MovieStartTimeValue,'String', num2str(MovieTimes(1),'%0.2f'))
        set(MovieEndTimeValue,'String', num2str(MovieTimes(2),'%0.2f'))
        UpdateDuration();
    end 

% -=- Callback: When the user changes file type -=-
    function inputFromFileTypeMenu(~,~)
        FileType = FileTypeList{get(FileTypeMenu,'value')};
        [~,MovName,~]=fileparts(MovName);
        switch FileType
            case 'Motion JPEG AVI'
                MovName = [MovName '.avi'];
            case 'MPEG-4'
                MovName = [MovName '.mp4'];
            case 'Archival'
                MovName = [MovName '.mj2'];
        end
        set(MovieDialog,'name',['Saving ' MovName ' ...']);
    end

% -=- Callback: When the user changes movie start/end time value boxes -=-
    function inputFromMovieStartTimeValue(~,~)
        limtmp=str2double(get(MovieStartTimeValue,'string')); 
        if isnan(limtmp)
            set(MovieStartTimeValue,'string',num2str(MovieTimes(1),'%0.2f'))
            return
        elseif limtmp<t(1)
            MovieTimes(1) = t(1);
        elseif limtmp>MovieTimes(2)
            MovieTimes(1) = MovieTimes(2);
        else
            MovieTimes(1) = limtmp;
        end
        set(MovieStartTimeValue,'string',num2str(MovieTimes(1),'%0.2f'))
        allSlides = get(MovieDialogSlider, 'Children');
        loc = get(allSlides(1),'Position');
        loc(1) = (MovieTimes(1) - t(1)) ...
            * (1-loc(3)) / (max(t)-min(t));
        set(allSlides(1),'position',loc)
        UpdateDuration();
    end 
    function inputFromMovieEndTimeValue(~,~)
        limtmp=str2double(get(MovieEndTimeValue,'string')); 
        if isnan(limtmp)
            set(MovieEndTimeValue,'string',num2str(MovieTimes(2),'%0.2f'))
            return
        elseif limtmp>t(end)
            MovieTimes(2) = t(end);
        elseif limtmp<MovieTimes(1)
            MovieTimes(2) = MovieTimes(1);
        else
            MovieTimes(2) = limtmp;
        end
        set(MovieEndTimeValue,'string',num2str(MovieTimes(2),'%0.2f'))
        allSlides = get(MovieDialogSlider, 'Children');
        loc = get(allSlides(2),'Position');
        loc(1) = (MovieTimes(2) - t(1)) ...
            * (1-loc(3)) / (max(t)-min(t));
        set(allSlides(2),'position',loc)
        UpdateDuration();
    end 

% -=- Callback: When the user changes new frame rate -=-
    function inputFromNewFrameRateValue(~,~)
        tmp = str2double(get(NewFrameRateValue,'string')); 
        if isnan(tmp) || tmp<=0.5 || isinf(tmp)
            set(NewFrameRateValue,'string',num2str(FrameRate,'%0.3f')); 
            return
        end
        FrameRate = tmp;
        UpdateDuration();
    end

% -=- Callback: When the user changes downsample rate -=-
    function inputFromDownsampleValue(~,~)
        tmp = str2double(get(DownsampleValue,'string')); 
        if isnan(tmp) || tmp<0
            set(DownsampleValue,'string',num2str(DownSample)); 
            return
        end
        DownSample = round(tmp);
        set(DownsampleValue,'string',num2str(DownSample)); 
        UpdateDuration();
    end

% -=- Update the duration of the movie to be saved -=-
    function UpdateDuration(~,~)
        set(DurationMessage,'string', ...
            ['new duration: ' ...
                num2str(diff(MovieTimes)/DownSample*fps/FrameRate, ...
                '%0.2f') ' s'])
    end

% -=- Callback: When the user clicks OK on the movie dialog -=-
    function inputFromMovieOKButton(~,~)
        close(MovieDialog)
        vid = VideoWriter(MovName,FileType);
        vid.FrameRate=FrameRate;
        open(vid);
        set(hf,'units','pixels');
        ind = find(t>=MovieTimes(1),1,'first') : ...
            DownSample : ...
            find(t<MovieTimes(2),1,'last');
        TimeList = t(ind);
        for jj=1:numel(TimeList)
            t1=TimeList(jj);
            changeTime()
            pause(1/FrameRate);
            hf_pos = get(hf,'position');
            if get(AxesButton,'value') % axes only
                snap=getframe(ha);
            else
                snap=getframe(hf,[AxesLeftPos AxesBottomPos ...
                    hf_pos(3)-AxesLeftPos hf_pos(4)-AxesBottomPos]);
            end
            writeVideo(vid,snap.cdata);
        end % for jj=1:numel(TimeList)
        close(vid);
        msgbox(['Saved ' MovName '.'])
    end

% -=- Callback: When the user clicks Cancel on the movie dialog -=-
    function inputFromMovieCancelButton(~,~)
        close(MovieDialog)
    end

    end % function Play(~)

% -=- Pause playback -=-
    function Pause(~)
        if isempty(get(MovNameText,'string'))
            set(PlayPauseButton,'string','play')
        else
            set(PlayPauseButton,'string','save...')
        end
        set([TimeText MovNameText FrameRateText OverlayButton ...
            TimeUnitText],'enable','on'); % re-enable controls
        set(TimeSlider,'callback',@inputFromTimeSlider); % re-enable slider
        t1=t(indt);
        changeTime() % just to clean up
        return
    end % function Pause(~)

% -=- Update controls and images when time changes -=-
    function changeTime(~)
        [~,indt]=min(abs(t1-t));
        indt=min([max([indt 1]) Nt]); % don't allow times outside the range
        set(TimeText,'string',num2str(t(indt),'%.5g')) % update the text box
        set(TimeSlider,'Value',t(indt)); % update slider
        set(ht,'string',['t = ' num2str(t(indt),'%0.2f') ' ' ...
            get(TimeUnitText,'string')])
        updateImageAndTracks();
        set(hf,'name',['t = ' num2str(t(indt),'%0.2f') ' ' ...
            get(TimeUnitText,'string') ': ' num2str(Np) ' particles'])
    end % function changeTime(~)

% -=- Update image and tracks -=-
    function updateImageAndTracks(~)

    % -=- Update image -=-
    switch movType
    case 'VideoReader'
        im = read(vobj,fr(indt));
    case 'seq'
        im = read_seq(fnames,fr(indt),0);
    case 'raw'
        im1=read_raw(fnames,fr(indt),[],0);
        if Nch==1
            im=im1{1};
        else
            im=repmat(0*im1{1}(:,:,1),[1 1 3]); % initialize
            for jj=1:Nch
                im(:,:,jj)=im1{jj};
            end
        end
    case 'stack'
        im = imread(fullfile(filepath,names.name),fr(indt));
    case 'images'
        im = imread(fullfile(filepath,names(fr(indt)).name));
    end % switch movType
    if ~strcmpi(movType,'none')
        if ~isRGB
            im = repmat(im,[1 1 3]); % force to grayscale
        end
        imscaled = double(im);
        imscaled(:,:,1) = (imscaled(:,:,1)-clim_ch1(1)) / diff(clim_ch1); % scale brightness
        imscaled(:,:,2) = (imscaled(:,:,2)-clim_ch2(1)) / diff(clim_ch2);
        imscaled(:,:,3) = (imscaled(:,:,3)-clim_ch3(1)) / diff(clim_ch3);
        imscaled(imscaled<0) = 0;
        set(hi,'cdata',imscaled) % update image
    end % if ~strcmpi(movType,'none')

    % -=- Update tracks -=-
    ind = find( (tr_fr_range(:,1)<=fr(indt)) ... % active tracks
        & (tr_fr_range(:,2)>=fr(indt)) );
    Np = numel(ind); % count of active tracks
    N_fr_show = fr(indt)-tr_fr_range(ind,1)+1; % for active each track, number of points to be plotted
    offsets = [0 ; cumsum(N_fr_show)];
    Npp = offsets(end)+Np; % total number of points to be plotted
    X1 = NaN(Npp,1);
    Y1 = NaN(Npp,1);
    cqty = ColorMenu.String;
    cqty = cqty{ColorMenu.Value};
    if strcmpi(cqty,'none')
        allcolors = lines(7);
        C1 = NaN(Npp,3);
    else
        C1 = NaN(Npp,1);
    end
    for jj=1:Np
        ind_tr = vtracks(ind(jj)).T<=fr(indt); % plot from beginning of track to current frame
        ind_off = jj+(offsets(jj):offsets(jj+1)-1); % where in the big vector to put the data
        X1(ind_off) = vtracks(ind(jj)).X(ind_tr)*umperpix;
        Y1(ind_off) = vtracks(ind(jj)).Y(ind_tr)*umperpix;
        switch cqty
            case 'instantaneous speed'
                C1(ind_off) = sqrt(vtracks(ind(jj)).U(ind_tr).^2 ...
                    + vtracks(ind(jj)).V(ind_tr).^2 )*umperpix*fps;
            case 'mean speed'
                C1(ind_off) = mean(sqrt( vtracks(ind(jj)).U.^2 ...
                    + vtracks(ind(jj)).V.^2 ))*umperpix*fps;
            case 'downstream velocity'
                C1(ind_off) = vtracks(ind(jj)).Udown(ind_tr)*umperpix*fps;
            case 'cross-stream velocity'
                C1(ind_off) = vtracks(ind(jj)).Ucross(ind_tr)*umperpix*fps;
            case 'none'
                C1(ind_off,:) = repmat(allcolors(mod(ind(jj),7)+1,:), ...
                    [sum(ind_tr),1]);
            case 'time'
                C1(ind_off) = vtracks(ind(jj)).T(ind_tr)/fps;
        end % switch cqty{ColorMenu.Value}
    end % for jj=1:Np
    if strcmpi(cqty,'none')
        set(hp,'edgecolor','flat','facevertexcdata',C1, ...
            'xdata',X1,'ydata',Y1);
        set(hcb,'visible','off')
    else
        set(hp,'edgecolor','interp','cdata',C1, ...
            'xdata',X1,'ydata',Y1);
        set(hcb,'visible','on')
    end % if strcmpi(cqty,'none')
    hcb.Label.String = cqty;

    end % function updateImageAndTracks(~)

end % function TracksPlotter
