function ha = plot_tracks(vtracks,bg,framerange,umperpix,fps,cqty,satlim)
% Usage: ha = plot_tracks(vtracks,[bg],[framerange],[umperpix],[fps],[cqty],[satlim])
% 
% Given particle tracks contained in "vtracks", a struct array of the sort
% produced by PredictiveTracker.m, plot_tracks plots the tracks. If a
% background image is provided in the array "bg", tracks are overlaid on
% the background. If a two-element vector "framerange" is provided,
% observations from frames before framerange(1) or after framerange(2) are
% not plotted. If the spatial resolution in microns per pixel is provided
% in "umperpix", all lengths are scaled accordingly. If the temporal
% resolution in frames per second is provided in "fps", all times are
% scaled accordingly. The color of the tracks is controlled by "cqty",
% which can contain one of these strings:
%   'speed': (default) tracks are color-coded according to instantaneous
%   speed;
%   'meanspeed': tracks are color-coded according to their mean speeds
%   'time': tracks are color-coded according to time
%   'Udown': tracks are color-coded according to instantaneous downstream
%   velocity;
%   'Ucross': tracks are color-coded according to instantaneous
%   cross-stream velocity;
%   'none': tracks are plotted as solid colors.
% To use 'Udown' or 'Ucross', provide "vtracks" in the form produced by
% calc_udotumeanhat.m. If "satlim" is provided, brightness of the
% background image is adjusted accordingly (e.g. give a value of 0.01 for
% 1% of pixels to be saturated). The axes handle is returned in "ha". 

% Next: Allow user to adjust contrast of background image with sliders.

% Written by Jeff Tithof.
% 24 May 2022: Updated by Doug Kelley to color-code tracks, allow for
% 3-channel background image, and plot on current axes (enabling use as a
% helper function). 
% 18 June 2022: dhk made satlim an input; default behavior is not to change
% the background image. 

TrackLineWidth = 2; % for plotting
framerange_default = [-inf inf]; % default range of times
umperpix_default = 1; % pixels
fps_default = 1; % frames
cqty_default = 'speed';

% Parse inputs
if nargin < 1
    error(['Usage: ha = ' mfilename ...
        '(vtracks,[bg],[framerange],[umperpix],[fps],[cqty],[satlim])'])
end
if ~exist('framerange','var') || isempty(framerange)
    framerange = framerange_default;
end
if ~exist('umperpix','var') || isempty(umperpix)
    umperpix = umperpix_default;
end
if ~exist('fps','var') || isempty(fps)
    fps = fps_default;
end
if ~exist('cqty','var') || isempty(cqty)
    cqty = cqty_default;
elseif ~strcmpi(cqty,'Udown') && ~strcmpi(cqty,'Ucross') ...
    && ~strcmpi(cqty,'speed') && ~strcmpi(cqty,'none') ...
    && ~strcmpi(cqty,'time') && ~strcmpi(cqty,'meanspeed')
    error('Uncrecognized string given for ''cqty''.')
end

% Plot tracks
Ntr = numel(vtracks);

for ii=1:Ntr % trim all tracks according to framerange
    ind = vtracks(ii).T >= framerange(1) & vtracks(ii).T <= framerange(2);
    vtracks(ii).len = sum(ind);
    vtracks(ii).X = vtracks(ii).X(ind);
    vtracks(ii).Y = vtracks(ii).Y(ind);
    vtracks(ii).T = vtracks(ii).T(ind);
    switch cqty
        case {'speed','meanspeed'}
            vtracks(ii).U = vtracks(ii).U(ind);
            vtracks(ii).V = vtracks(ii).V(ind);
        case 'Udown'
            vtracks(ii).Udown = vtracks(ii).Udown(ind);
        case 'Ucross'
            vtracks(ii).Ucross = vtracks(ii).Ucross(ind);
    end
end
vtracks( [vtracks.len] == 0) = []; % delete empty tracks
Ntr = numel(vtracks); % updated track count
Np = sum([vtracks.len]); % count of all particles in all tracks
X1 = NaN(Np+Ntr,1);
Y1 = NaN(Np+Ntr,1);
if strcmpi(cqty,'none')
    C1 = NaN(Np+Ntr,3);
    allcolors = lines(Ntr);
else
    C1 = NaN(Np+Ntr,1);
end
offsets = cumsum([vtracks.len]); 
offsets = [0 offsets]; 
for ii=1:Ntr % concatenate all positions and speeds
    ind = ii+(offsets(ii):offsets(ii+1)-1);
    X1(ind) = vtracks(ii).X*umperpix;
    Y1(ind) = vtracks(ii).Y*umperpix;
    switch cqty
        case 'speed'
            C1(ind) = sqrt( vtracks(ii).U.^2 ...
                + vtracks(ii).V.^2 )*umperpix*fps;
        case 'meanspeed'
            C1(ind) = mean(sqrt( vtracks(ii).U.^2 ...
                + vtracks(ii).V.^2 ))*umperpix*fps;
        case 'Udown'
            C1(ind) = vtracks(ii).Udown*umperpix*fps;
        case 'Ucross'
            C1(ind) = vtracks(ii).Ucross*umperpix*fps;
        case 'none'
            C1(ind,:) = repmat(allcolors(ii,:),[numel(ind),1]);
        case 'time'
            C1(ind) = vtracks(ii).T/fps;
    end
end
hp = patch(X1,Y1,C1(:,1),'edgecolor','interp','linewidth',TrackLineWidth);
if strcmpi(cqty,'none')
    set(hp,'edgecolor','flat','facevertexcdata',C1);
end
set(gca,'dataaspectratio',[1 1 1],'nextplot','add','layer','top', ...
    'box','on','ydir','reverse') % reverse ydir per the norm for images; this data comes from images!

% Plot background image, adjusting contrast
if exist('bg','var') && ~isempty(bg)
    if size(bg,3) == 1 % grayscale
        bg = repmat(bg,[1 1 3]);
    end
    bg = double(bg);
    if exist('satlim','var') && ~isempty(satlim)
        for ii=1:3
            pixvals=sort(bg(:,:,ii),'descend');
            ind=round(satlim*numel(pixvals));
            bg(:,:,ii) = bg(:,:,ii)/double(pixvals(ind))*65535;
        end
        bg=uint16(bg);
    end
    [ht,wd,~] = size(bg);
    hi = imagesc((0:wd-1)*umperpix,(0:ht-1)*umperpix,bg);
    uistack(hi,'bottom')
end
axis tight

if nargout > 0
    ha = gca;
end

end
